from __future__ import annotations

import abc

import numpy as np
from scipy.interpolate import interp1d
from scipy.spatial.transform import Rotation as R

from ... import config
from ...config import types


def _reshape_vectors(rot_pts, x, y, z):
    xrot = rot_pts[:, 0].reshape(x.shape)
    yrot = rot_pts[:, 1].reshape(y.shape)
    zrot = rot_pts[:, 2].reshape(z.shape)

    return xrot, yrot, zrot


def rotate_3d_mesh_by_vec(
    x: np.ndarray, y: np.ndarray, z: np.ndarray, rot_versor: np.ndarray, angle: float
):
    """
    Rotate meshgrid points according to a rotation versor and angle.

    :param numpy.ndarray[numpy.ndarray[float]] x: x coordinate points of the meshgrid
    :param numpy.ndarray[numpy.ndarray[float]] y: y coordinate points of the meshgrid
    :param numpy.ndarray[numpy.ndarray[float]] z: z coordinate points of the meshgrid
    :param numpy.ndarray[float] rot_versor: vector representing rotation versor
    :param float angle: rotation angle in radian
    :return: Rotated x, y, z coordinate points
    :rtype: tuple[numpy.ndarray[numpy.ndarray[float]]
    """

    # Arrange point coordinates in shape (N, 3) for vectorized processing
    pts = np.array([x.ravel(), y.ravel(), z.ravel()]).transpose()

    # Create and apply rotation
    rot = R.from_rotvec(rot_versor * angle)
    rot_pts = rot.apply(pts)

    # return to original shape of meshgrid
    return _reshape_vectors(rot_pts, x, y, z)


def translate_3d_mesh_by_vec(
    x: np.ndarray, y: np.ndarray, z: np.ndarray, t_vec: np.ndarray
):
    """
    Translate meshgrid points according to a 3d vector.

    :param numpy.ndarray[numpy.ndarray[float]] x: x coordinate points of the meshgrid
    :param numpy.ndarray[numpy.ndarray[float]] y: y coordinate points of the meshgrid
    :param numpy.ndarray[numpy.ndarray[float]] z: z coordinate points of the meshgrid
    :param numpy.ndarray[float] t_vec: translation vector
    :return: Translated x, y, z coordinate points
    :rtype: tuple[numpy.ndarray[numpy.ndarray[float]]
    """

    # Arrange point coordinates in shape (N, 3) for vectorized processing
    pts = np.array([x.ravel(), y.ravel(), z.ravel()]).transpose()
    pts = pts + t_vec
    # return to original shape of meshgrid
    return _reshape_vectors(pts, x, y, z)


def rotate_3d_mesh_by_rot_mat(
    x: np.ndarray, y: np.ndarray, z: np.ndarray, rot_mat: np.ndarray
):
    """
    Rotate meshgrid points according to a rotation matrix.

    :param numpy.ndarray[numpy.ndarray[float]] x: x coordinate points of the meshgrid
    :param numpy.ndarray[numpy.ndarray[float]] y: y coordinate points of the meshgrid
    :param numpy.ndarray[numpy.ndarray[float]] z: z coordinate points of the meshgrid
    :param numpy.ndarray[numpy.ndarray[float]] rot_mat: rotation matrix, shape (3,3)
    :return: Rotated x, y, z coordinate points
    :rtype: tuple[numpy.ndarray[numpy.ndarray[float]]
    """

    # Arrange point coordinates in shape (N, 3) for vectorized processing
    pts = np.array([x.ravel(), y.ravel(), z.ravel()]).transpose()

    # Create and apply rotation
    rot = R.from_matrix(rot_mat)
    rot_pts = rot.apply(pts)

    # return to original shape of meshgrid
    return _reshape_vectors(rot_pts, x, y, z)


def _surface_resampling(
    surface_function,
    theta_min=0,
    theta_max=2 * np.pi,
    phi_min=0,
    phi_max=np.pi,
    precision=25,
):
    # first sampling to estimate surface distribution
    theta, phi = np.meshgrid(
        np.linspace(theta_min, theta_max, precision),
        np.linspace(phi_min, phi_max, precision),
    )
    coords = surface_function(theta, phi)

    # estimate surfaces, decomposing it into parallelograms along theta and phi
    delta_t_temp = np.diff(coords, axis=2)
    delta_u_temp = np.diff(coords, axis=1)
    delta_t = np.zeros(coords.shape)
    delta_u = np.zeros(coords.shape)
    delta_t[: coords.shape[0], : coords.shape[1], 1 : coords.shape[2]] = delta_t_temp
    delta_u[: coords.shape[0], 1 : coords.shape[1], : coords.shape[2]] = delta_u_temp
    delta_S = np.linalg.norm(np.cross(delta_t, delta_u, 0, 0), axis=2)
    cum_S_t = np.cumsum(delta_S.sum(axis=0))
    cum_S_u = np.cumsum(delta_S.sum(axis=1))
    return theta, phi, cum_S_t, cum_S_u


def uniform_surface_sampling(
    n_points,
    surface_function,
    theta_min=0,
    theta_max=2 * np.pi,
    phi_min=0,
    phi_max=np.pi,
    precision=25,
):
    """
    Uniform-like random sampling of polar coordinates based on surface estimation.
    This sampling is useful on elliptic surfaces (e.g. sphere).
    Algorithm based on https://github.com/maxkapur/param_tools

    :param int n_points: number of points to sample
    :param Callable[..., numpy.ndarray[float]] surface_function: function converting polar
        coordinates into cartesian coordinates
    :param int precision: size of grid used to estimate function surface
    """

    theta, phi, cum_S_t, cum_S_u = _surface_resampling(
        surface_function, theta_min, theta_max, phi_min, phi_max, precision
    )
    # resample along the cumulative surface to uniformize point distribution
    # equivalent to a multinomial sampling
    sampled_t = np.random.rand(n_points) * cum_S_t[-1]
    sampled_u = np.random.rand(n_points) * cum_S_u[-1]
    sampled_t = interp1d(cum_S_t, theta[0, :])(sampled_t)
    sampled_u = interp1d(cum_S_u, phi[:, 0])(sampled_u)

    return surface_function(sampled_t, sampled_u)


def uniform_surface_wireframe(
    n_points_1,
    n_points_2,
    surface_function,
    theta_min=0,
    theta_max=2 * np.pi,
    phi_min=0,
    phi_max=np.pi,
    precision=25,
):
    """
    Uniform-like meshgrid of size (n_point_1, n_points_2) of polar coordinates based on
    surface estimation.
    This meshgrid is useful on elliptic surfaces (e.g. sphere).
    Algorithm based on https://github.com/maxkapur/param_tools

    :param Callable[..., numpy.ndarray[float]] surface_function: function converting polar
        coordinates into cartesian coordinates
    :param int precision: size of grid used to estimate function surface
    """

    theta, phi, cum_S_t, cum_S_u = _surface_resampling(
        surface_function, theta_min, theta_max, phi_min, phi_max, precision
    )
    sampled_t = np.linspace(0, cum_S_t[-1], n_points_1)
    sampled_u = np.linspace(0, cum_S_u[-1], n_points_2)
    sampled_t = interp1d(cum_S_t, theta[0, :])(sampled_t)
    sampled_u = interp1d(cum_S_u, phi[:, 0])(sampled_u)
    sampled_t, sampled_u = np.meshgrid(sampled_t, sampled_u)
    return surface_function(sampled_t, sampled_u)


def _get_prod_angle_vector(hv, z_versor=None):
    """
    Calculate the cross product and the arc cosines angle between two vectors.

    :param numpy.ndarray hv: vector to rotate
    :param numpy.ndarray z_versor: reference vector
    :return: cross product and arc cosines angle
    :rtype: tuple[numpy.ndarray, numpy.ndarray]
    """

    if z_versor is None:
        z_versor = np.array([0, 0, 1])
    return np.cross(z_versor, hv), np.arccos(np.dot(hv, z_versor))


def _get_rotation_vector(hv, z_versor=None, positive_angle=True):
    """
    Calculate the rotation vector between two vectors.

    :param numpy.ndarray hv: vector to rotate
    :param numpy.ndarray z_versor: reference vector
    :param bool positive_angle: if False, the angle is inverted
    :return: rotation vector
    :rtype: scipy.spatial.transform.Rotation
    """

    if z_versor is None:
        z_versor = np.array([0, 0, 1])
    perp, angle = _get_prod_angle_vector(hv, z_versor)
    angle = angle if positive_angle else -angle
    rot = R.from_rotvec(perp * angle)
    return rot


def _rotate_by_coord(x, y, z, hv, origin, test_hv=False):
    perp, angle = _get_prod_angle_vector(hv / np.linalg.norm(hv))

    x, y, z = rotate_3d_mesh_by_vec(x, y, z, perp, angle)
    if test_hv and hv[2] < 0:
        z = -z
    return translate_3d_mesh_by_vec(x, y, z, origin)


def _get_extrema_after_rot(extrema, origin, top_center):
    height = np.linalg.norm(top_center - origin)
    rot = _get_rotation_vector((top_center - origin) / height)

    for i, pt in enumerate(extrema):
        extrema[i] = rot.apply(pt)

    return np.min(extrema, axis=0) + origin, np.max(extrema, axis=0) + origin


def inside_mbox(
    points: np.ndarray[float],
    mbb_min: np.ndarray[float],
    mbb_max: np.ndarray[float],
):
    """
    Check if the points given in input are inside the minimal bounding box.

    :param numpy.ndarray points: An array of 3D points.
    :param numpy.ndarray mbb_min: 3D point representing the lowest coordinate of the
        minimal bounding box.
    :param numpy.ndarray mbb_max: 3D point representing the highest coordinate of the
        minimal bounding box.
    :return: A bool np.ndarray specifying whether each point of the input array is inside
        the minimal bounding box or not.
    :rtype: numpy.ndarray[bool]
    """
    inside = (
        (points[:, 0] > mbb_min[0])
        & (points[:, 0] < mbb_max[0])
        & (points[:, 1] > mbb_min[1])
        & (points[:, 1] < mbb_max[1])
        & (points[:, 2] > mbb_min[2])
        & (points[:, 2] < mbb_max[2])
    )

    return inside


@config.dynamic(attr_name="type", default="shape", auto_classmap=True)
class GeometricShape(abc.ABC):
    """
    Base class for geometric shapes.
    """

    epsilon = config.attr(type=float, required=False, default=1.0e-3)
    """
    Tolerance value to compare coordinates.
    """

    def __init__(self, **kwargs):
        self.mbb_min, self.mbb_max = self.find_mbb()

    @abc.abstractmethod
    def find_mbb(self):  # pragma: no cover
        """
        Compute the minimum bounding box surrounding the shape.
        """
        pass

    def check_mbox(self, points: np.ndarray[float]):
        """
        Check if the points given in input are inside the minimal bounding box.

        :param numpy.ndarray points: A cloud of points.
        :return: A bool np.ndarray specifying whether each point of the input array is
            inside the minimal bounding box or not.
        :rtype: numpy.ndarray
        """
        return inside_mbox(points, self.mbb_min, self.mbb_max)

    @abc.abstractmethod
    def get_volume(self):  # pragma: no cover
        """
        Get the volume of the geometric shape.

        :return: The volume of the geometric shape.
        :rtype: float
        """
        pass

    @abc.abstractmethod
    def translate(self, t_vector: np.ndarray[float]):  # pragma: no cover
        """
        Translate the geometric shape by the vector t_vector.

        :param numpy.ndarray t_vector: The displacement vector
        """
        pass

    @abc.abstractmethod
    def rotate(self, r_versor: np.ndarray[float], angle: float):  # pragma: no cover
        """
        Rotate all the shapes around r_versor, which is a versor passing through the
        origin, by the specified angle.

        :param r_versor: A versor specifying the rotation axis.
        :type r_versor: numpy.ndarray[float]
        :param float angle: the rotation angle, in radians.
        """
        pass

    @abc.abstractmethod
    def generate_point_cloud(self, npoints: int):  # pragma: no cover
        """
        Generate a point cloud made by npoints points.

        :param int npoints: The number of points to generate.
        :return: a (npoints x 3) numpy array.
        :rtype: numpy.ndarray
        """
        pass

    @abc.abstractmethod
    def check_inside(
        self, points: np.ndarray[float]
    ) -> np.ndarray[bool]:  # pragma: no cover
        """
        Check if the points given in input are inside the geometric shape.

        :param numpy.ndarray points: A cloud of points.
        :return: A bool array with same length as points, containing whether the -ith
            point is inside the geometric shape or not.
        :rtype: numpy.ndarray
        """
        pass

    @abc.abstractmethod
    def wireframe_points(self, nb_points_1=30, nb_points_2=30):  # pragma: no cover
        """
        Generate a wireframe to plot the geometric shape. If a sampling of points is
        needed (e.g. for sphere), the wireframe is based on a grid of shape (nb_points_1,
        nb_points_2).

        :param int nb_points_1: number of points sampled along the first dimension
        :param int nb_points_2: number of points sampled along the second dimension
        :return: Coordinate components of the wireframe
        :rtype: tuple[numpy.ndarray[numpy.ndarray[float]]
        """
        pass


@config.node
class ShapesComposition:
    """
    A collection of geometric shapes, which can be labelled to distinguish different parts
    of a neuron.
    """

    shapes = config.list(
        type=GeometricShape,
        required=types.same_size("shapes", "labels", required=True),
        hint=[{"type": "sphere", "radius": 40.0, "center": [0.0, 0.0, 0.0]}],
    )
    """
    List of GeometricShape that make up the neuron.
    """
    labels = config.list(
        type=types.list(),
        required=types.same_size("shapes", "labels", required=True),
        hint=[["soma", "dendrites", "axon"]],
    )
    """
    List of lists of labels associated to each geometric shape.
    """
    voxel_size = config.attr(type=float, required=False, default=1.0)
    """
    Dimension of the side of a voxel, used to determine how many points must be generated
    to represent the geometric shape.
    """

    def __init__(self, **kwargs):
        # The two corners individuating the minimal bounding box.
        self._mbb_min = np.array([0.0, 0.0, 0.0])
        self._mbb_max = np.array([0.0, 0.0, 0.0])

        self.find_mbb()

    def add_shape(self, shape: GeometricShape, labels: list[str]):
        """
        Add a geometric shape to the collection.

        :param GeometricShape shape: A GeometricShape to add to the collection.
        :param list[str] labels: A list of labels for the geometric shape to add.
        """
        # Update mbb
        if len(self._shapes) == 0:
            self._mbb_min = np.copy(shape.mbb_min)
            self._mbb_max = np.copy(shape.mbb_max)
        else:
            self._mbb_min = np.minimum(self._mbb_min, shape.mbb_min)
            self._mbb_max = np.maximum(self._mbb_max, shape.mbb_max)
        self._shapes.append(shape)
        self._labels.append(labels)

    def filter_by_labels(self, labels: list[str]) -> ShapesComposition:
        """
        Filter the collection of shapes, returning only the ones corresponding to the
        given labels.

        :param list[str] labels: A list of labels.
        :return: A new ShapesComposition object containing only the shapes labelled as
            specified.
        :rtype: ShapesComposition
        """
        result = ShapesComposition(dict(voxel_size=self.voxel_size, labels=[], shapes=[]))
        selected_id = np.where(np.isin(labels, self._labels))[0]
        result._shapes = [self._shapes[i].__copy__() for i in selected_id]
        result._labels = [self._labels[i].copy() for i in selected_id]
        result.mbb_min, result.mbb_max = result.find_mbb()
        return result

    def translate(self, t_vec: np.ndarray[float]):
        """
        Translate all the shapes in the collection by the vector t_vec. It also
        automatically translate the minimal bounding box.

        :param numpy.ndarray t_vec: The displacement vector.
        """
        for shape in self._shapes:
            shape.translate(t_vec)
        self._mbb_min += t_vec
        self._mbb_max += t_vec

    def get_volumes(self) -> list[float]:
        """
        Compute the volumes of all the shapes.

        :rtype: list[float]
        """
        return [shape.get_volume() for shape in self._shapes]

    def get_mbb_min(self):
        """
        Returns the bottom corner of the minimum bounding box containing the collection of
        shapes.

        :return: The bottom corner individuating the minimal bounding box of the shapes
            collection.
        :rtype: numpy.ndarray[float]
        """
        return self._mbb_min

    def get_mbb_max(self):
        """
        Returns the top corner of the minimum bounding box containing the collection of
        shapes.

        :return: The top corner individuating the minimal bounding box of the shapes
            collection.
        :rtype: numpy.ndarray[float]
        """
        return self._mbb_max

    def find_mbb(self) -> tuple[np.ndarray[float], np.ndarray[float]]:
        """
        Compute the minimal bounding box containing the collection of shapes.

        :return: The two corners individuating the minimal bounding box of the shapes
            collection.
        :rtype: tuple(numpy.ndarray[float], numpy.ndarray[float])
        """
        mins = np.empty([len(self._shapes), 3])
        maxs = np.empty([len(self._shapes), 3])
        for i, shape in enumerate(self._shapes):
            mins[i, :] = shape.mbb_min
            maxs[i, :] = shape.mbb_max
        self._mbb_min = np.min(mins, axis=0) if len(self._shapes) > 0 else np.zeros(3)
        self._mbb_max = np.max(maxs, axis=0) if len(self._shapes) > 0 else np.zeros(3)
        return self._mbb_min, self._mbb_max

    def compute_n_points(self) -> list[int]:
        """
        Compute the number of points to generate in a point cloud, using the dimension of
        the voxel specified in self._voxel_size.

        :return: The number of points to generate.
        :rtype: numpy.ndarray[int]
        """
        return [int(shape.get_volume() // self.voxel_size**3) for shape in self._shapes]

    def generate_point_cloud(self) -> np.ndarray[float] | None:
        """
        Generate a point cloud. The number of points to generate is determined
        automatically using the voxel size.

        :return: A numpy.ndarray containing the 3D points of the cloud. If there are no
            shapes in the collection, it returns None.
        :rtype: numpy.ndarray[float] | None
        """
        if len(self._shapes) != 0:
            return np.concatenate(
                [
                    shape.generate_point_cloud(numpts)
                    for shape, numpts in zip(
                        self._shapes, self.compute_n_points(), strict=False
                    )
                ]
            )
        else:
            return None

    def generate_wireframe(
        self,
        nb_points_1=30,
        nb_points_2=30,
    ) -> tuple[list, list, list] | None:
        """
        Generate the wireframes of a collection of shapes. If a sampling of points is
        needed for certain shapes (e.g. for sphere), their wireframe is based on a grid of
        shape (nb_points_1, nb_points_2).

        :param int nb_points_1: number of points sampled along the first dimension
        :param int nb_points_2: number of points sampled along the second dimension
        :return: The x,y,z coordinates of the wireframe of each shape.
        :rtype: tuple[list[numpy.ndarray[numpy.ndarray[float]]]] | None
        """
        if len(self._shapes) != 0:
            x = []
            y = []
            z = []
            for shape in self._shapes:
                # For each shape, the shape of the wireframe is different, so we need to
                # append them manually
                xt, yt, zt = shape.wireframe_points(
                    nb_points_1=nb_points_1, nb_points_2=nb_points_2
                )
                x.append(xt)
                y.append(yt)
                z.append(zt)
            return x, y, z
        return None

    def inside_mbox(self, points: np.ndarray[float]) -> np.ndarray[bool]:
        """
        Check if the points given in input are inside the minimal bounding box of the
        collection.

        :param numpy.ndarray points: An array of 3D points.
        :return: A bool np.ndarray specifying whether each point of the input array is
            inside the minimal bounding box of the collection.
        :rtype: numpy.ndarray[bool]
        """
        return inside_mbox(points, self._mbb_min, self._mbb_max)

    def inside_shapes(self, points: np.ndarray[float]) -> np.ndarray[bool] | None:
        """
        Check if the points given in input are inside at least in one of the shapes of the
        collection.

        :param numpy.ndarray points: An array of 3D points.
        :return: A bool numpy.ndarray specifying whether each point of the input array is
            inside the collection of shapes or not.
        :rtype: numpy.ndarray[bool]
        """
        if len(self._shapes) != 0:
            is_inside = np.full(len(points), 0, dtype=bool)
            for shape in self._shapes:
                tmp = shape.check_mbox(points)
                if np.any(tmp):
                    is_inside = is_inside | shape.check_inside(points)
            return is_inside
        else:
            return None


@config.node
class Ellipsoid(GeometricShape, classmap_entry="ellipsoid"):
    """
    An ellipsoid, described in cartesian coordinates.
    """

    origin = config.attr(
        type=types.ndarray(shape=(3,), dtype=float), required=True, hint=[0.0, 0.0, 0.0]
    )
    """
    The coordinates of the center of the ellipsoid.
    """
    lambdas = config.attr(
        type=types.ndarray(shape=(3,), dtype=float), required=True, hint=[1.0, 0.5, 2.0]
    )
    """
    The length of the three semi-axes.
    """

    @config.property(type=types.ndarray(shape=(3,)), required=True)
    def v0(self):
        """
        The versor on which the first semi-axis lies.
        """
        return self._v0

    @v0.setter
    def v0(self, value):
        self._v0 = np.copy(value) / np.linalg.norm(value)

    @config.property(type=types.ndarray(shape=(3,)), required=True)
    def v1(self):
        """
        The versor on which the second semi-axis lies.
        """
        return self._v1

    @v1.setter
    def v1(self, value):
        self._v1 = np.copy(value) / np.linalg.norm(value)

    @config.property(type=types.ndarray(shape=(3,)), required=True)
    def v2(self):
        """
        The versor on which the third semi-axis lies.
        """
        return self._v2

    @v2.setter
    def v2(self, value):
        self._v2 = np.copy(value) / np.linalg.norm(value)

    def find_mbb(self):
        # Find the minimum bounding box, to avoid computing it every time
        extrema = (
            np.array(
                [
                    self.lambdas[0] * self.v0,
                    -self.lambdas[0] * self.v0,
                    self.lambdas[1] * self.v1,
                    -self.lambdas[1] * self.v1,
                    self.lambdas[2] * self.v2,
                    -self.lambdas[2] * self.v2,
                ]
            )
            + self.origin
        )
        mbb_min = np.min(extrema, axis=0)
        mbb_max = np.max(extrema, axis=0)
        return mbb_min, mbb_max

    def get_volume(self):
        return np.pi * self.lambdas[0] * self.lambdas[1] * self.lambdas[2]

    def translate(self, t_vector: np.ndarray):
        self.origin += t_vector
        self.mbb_min += t_vector
        self.mbb_max += t_vector

    def rotate(self, r_versor: np.ndarray[float], angle: float):
        rot = R.from_rotvec(r_versor * angle)
        self.v0 = rot.apply(self.v0)
        self.v1 = rot.apply(self.v1)
        self.v2 = rot.apply(self.v2)

    def surface_point(self, theta, phi):
        """
        Convert polar coordinates into their 3D location on the ellipsoid surface.

        :param float|numpy.ndarray[float] theta: first polar coordinate in [0; 2*np.pi]
        :param float|numpy.ndarray[float] phi: second polar coordinate in [0; np.pi]
        :return: surface coordinates
        :rtype: float|numpy.ndarray[float]
        """
        return np.array(
            [
                self.lambdas[0] * np.cos(theta) * np.sin(phi),
                self.lambdas[1] * np.sin(theta) * np.sin(phi),
                self.lambdas[2] * np.cos(phi),
            ]
        )

    def generate_point_cloud(self, npoints: int):
        sampling = uniform_surface_sampling(npoints, self.surface_point)
        sampling = sampling.T * np.random.rand(npoints, 3)  # sample within the shape

        # Rotate the ellipse
        rmat = np.array([self.v0, self.v1, self.v2]).T
        sampling = sampling.dot(rmat)
        sampling = sampling + self.origin
        return sampling

    def check_inside(self, points: np.ndarray[float]):
        # Check if the quadratic form associated to the ellipse is less than 1 at a point
        diff = points - self.origin
        vmat = np.array([self.v0, self.v1, self.v2])
        diag = np.diag(1 / self.lambdas**2)
        qmat = vmat.dot(diag).dot(vmat)
        quad_prod = np.diagonal(diff.dot(qmat.dot(diff.T)))

        # Check if the points are inside the ellipsoid
        inside_points = quad_prod < 1
        return inside_points

    def wireframe_points(self, nb_points_1=30, nb_points_2=30):
        # Generate an ellipse orientated along x,y,z
        x, y, z = uniform_surface_wireframe(nb_points_1, nb_points_2, self.surface_point)
        # Rotate the ellipse
        rmat = np.array([self.v0, self.v1, self.v2]).T
        x, y, z = rotate_3d_mesh_by_rot_mat(x, y, z, rmat)
        return translate_3d_mesh_by_vec(x, y, z, self.origin)


@config.node
class Cone(GeometricShape, classmap_entry="cone"):
    """
    A cone, described in cartesian coordinates.
    """

    apex = config.attr(
        type=types.ndarray(shape=(3,), dtype=float), required=True, hint=[0.0, 1.0, 0.0]
    )
    """
    The coordinates of the apex of the cone.
    """
    origin = config.attr(
        type=types.ndarray(shape=(3,), dtype=float), required=True, hint=[0.0, 0.0, 0.0]
    )
    """
    The coordinates of the center of the cone's base.
    """
    radius = config.attr(type=float, required=False, default=1.0)
    """
    The radius of the base circle.
    """

    def find_mbb(self):
        # Vectors identifying half of the sides of the base rectangle in xy
        u = np.array([self.radius, 0, 0])
        v = np.array([0, self.radius, 0])

        # Find the rotation angle and axis
        hv = self.origin - self.apex
        rot = _get_rotation_vector(hv / np.linalg.norm(hv))

        # Rotated vectors of the box
        v1 = rot.apply(u)
        v2 = rot.apply(v)
        v3 = self.origin - self.apex

        # Coordinates identifying the minimal bounding box
        minima = np.min([v1, v2, v3, -v1, -v2], axis=0)
        maxima = np.max([v1, v2, v3, -v1, -v2], axis=0)
        return minima, maxima

    def get_volume(self):
        h = np.linalg.norm(self.apex - self.origin)
        b = np.pi * self.radius * self.radius
        return b * h / 3

    def translate(self, t_vector):
        self.origin += t_vector
        self.apex += t_vector
        self.mbb_min += t_vector
        self.mbb_max += t_vector

    def rotate(self, r_versor: np.ndarray[float], angle: float):
        rot = R.from_rotvec(r_versor * angle)
        self.apex = rot.apply(self.apex)

    def generate_point_cloud(self, npoints: int):
        theta = np.pi * 2.0 * np.random.rand(npoints)
        rand_a = np.random.rand(npoints)
        rand_b = np.random.rand(npoints)

        # Height vector
        hv = self.origin - self.apex
        cloud = np.full((npoints, 3), 0, dtype=float)

        # Generate a cone with the apex in the origin and the origin at (0,0,1)
        cloud[:, 0] = (self.radius * rand_a * np.cos(theta)) * rand_b
        cloud[:, 1] = self.radius * rand_a * np.sin(theta)
        cloud[:, 2] = rand_a * np.linalg.norm(hv)

        # Rotate the cone: Find the axis of rotation and compute the angle
        perp, angle = _get_prod_angle_vector(hv / np.linalg.norm(hv))

        if hv[2] < 0:
            cloud[:, 2] = -cloud[:, 2]
        rot = R.from_rotvec(perp * angle)
        cloud = rot.apply(cloud)

        # Translate the cone
        cloud = cloud + self.apex
        return cloud

    def check_inside(self, points: np.ndarray[float]):
        # Find the vector w of the height.
        h_vector = self.origin - self.apex
        height = np.linalg.norm(h_vector)
        h_vector /= height
        # Center the points
        pts = points - self.apex

        # Rotate back to xyz
        rot = _get_rotation_vector(h_vector)
        rot_pts = rot.apply(pts)

        # Find the angle between the points and the apex
        apex_angles = np.arccos(
            np.dot((rot_pts / np.linalg.norm(rot_pts, axis=1)[..., np.newaxis]), h_vector)
        )
        # Compute the cone angle
        cone_angle = np.arctan(self.radius / height)

        # Select the points inside the cone
        inside_points = (
            (apex_angles < cone_angle + self.epsilon)
            & (rot_pts[:, 2] > np.min([self.origin[2], self.apex[2]]) - self.epsilon)
            & (rot_pts[:, 2] < np.max([self.origin[2], self.apex[2]]) + self.epsilon)
        )
        return inside_points

    def wireframe_points(self, nb_points_1=30, nb_points_2=30):
        # Set up the grid in polar coordinates
        theta = np.linspace(0, 2 * np.pi, nb_points_1)
        r = np.linspace(0, self.radius, nb_points_2)
        theta, r = np.meshgrid(theta, r)

        # Height vector
        hv = np.array(self.origin) - np.array(self.apex)
        height = np.linalg.norm(hv)
        # angle = np.arctan(height/self.radius)

        # Generate a cone with the apex in the origin and the center at (0,0,1)
        x = r * np.cos(theta)
        y = r * np.sin(theta)
        z = r * height / self.radius

        # Rotate the cone
        return _rotate_by_coord(x, y, z, hv, self.apex, test_hv=True)


@config.node
class Cylinder(GeometricShape, classmap_entry="cylinder"):
    """
    A cylinder, described in cartesian coordinates.
    """

    origin = config.attr(
        type=types.ndarray(shape=(3,), dtype=float), required=True, hint=[0.0, 0.0, 0.0]
    )
    """
    The coordinates of the center of the bottom circle of the cylinder.
    """
    top_center = config.attr(
        type=types.ndarray(shape=(3,), dtype=float), required=True, hint=[0.0, 2.0, 0.0]
    )
    """
    The coordinates of the center of the top circle of the cylinder.
    """
    radius = config.attr(type=float, required=False, default=1.0)
    """
    The radius of the base circle.
    """

    def find_mbb(self):
        height = np.linalg.norm(self.top_center - self.origin)
        # Extrema of the xyz standard cyl
        extrema = [
            np.array([-self.radius, -self.radius, 0.0]),
            np.array([-self.radius, self.radius, 0.0]),
            np.array([self.radius, -self.radius, 0.0]),
            np.array([self.radius, self.radius, 0.0]),
            np.array([self.radius, self.radius, height]),
            np.array([-self.radius, self.radius, height]),
            np.array([self.radius, -self.radius, height]),
            np.array([-self.radius, -self.radius, height]),
        ]

        # Rotate the cylinder

        return _get_extrema_after_rot(extrema, self.origin, self.top_center)

    def get_volume(self):
        h = np.linalg.norm(self.top_center - self.origin)
        b = np.pi * self.radius * self.radius
        return b * h

    def translate(self, t_vector: np.ndarray[float]):
        self.origin += t_vector
        self.top_center += t_vector
        self.mbb_min += t_vector
        self.mbb_max += t_vector

    def rotate(self, r_versor: np.ndarray[float], angle: float):
        rot = R.from_rotvec(r_versor * angle)
        # rotation according to bottom center
        self.top_center = rot.apply(self.top_center)

    def generate_point_cloud(self, npoints: int):
        # Generate an ellipse orientated along x,y,z
        cloud = np.full((npoints, 3), 0, dtype=float)
        theta = np.pi * 2.0 * np.random.rand(npoints)
        rand = np.random.rand(npoints, 3)
        height = np.linalg.norm(self.top_center - self.origin)

        # Generate an ellipsoid centered at the origin, with the semiaxes on x,y,z
        cloud[:, 0] = self.radius * np.cos(theta)
        cloud[:, 1] = self.radius * np.sin(theta)
        cloud[:, 2] = height
        cloud = cloud * rand

        # Rotate the cylinder
        hv = (self.top_center - self.origin) / height
        rot = _get_rotation_vector(hv / np.linalg.norm(hv))
        cloud = rot.apply(cloud)
        cloud = cloud + self.origin
        return cloud

    def check_inside(self, points: np.ndarray[float]):
        # Translate back to origin
        pts = points - self.origin

        # Rotate back to xyz
        height = np.linalg.norm(self.top_center - self.origin)
        rot = _get_rotation_vector(
            (self.top_center - self.origin) / height, positive_angle=False
        )
        rot_pts = rot.apply(pts)
        # Check for intersections
        inside_points = (
            (rot_pts[:, 2] < height + self.epsilon)
            & (rot_pts[:, 2] > -self.epsilon)
            & (
                rot_pts[:, 0] * rot_pts[:, 0] + rot_pts[:, 1] * rot_pts[:, 1]
                < self.radius**2 + self.epsilon
            )
        )
        return inside_points

    def wireframe_points(self, nb_points_1=30, nb_points_2=30):
        # Set up the grid in polar coordinates
        theta = np.linspace(0, 2 * np.pi, nb_points_1)

        # Height vector
        hv = np.array(self.origin) - np.array(self.top_center)
        height = np.linalg.norm(hv)

        h = np.linspace(0, height, nb_points_2)
        theta, h = np.meshgrid(theta, h)

        x = self.radius * np.cos(theta)
        y = self.radius * np.sin(theta)
        z = h

        # Rotate the cylinder
        hv = (self.top_center - self.origin) / height
        return _rotate_by_coord(x, y, z, hv, self.origin)


@config.node
class Sphere(GeometricShape, classmap_entry="sphere"):
    """
    A sphere, described in cartesian coordinates.
    """

    origin = config.attr(
        type=types.ndarray(shape=(3,), dtype=float), required=True, hint=[0.0, 0.0, 0.0]
    )
    """
    The coordinates of the center of the sphere.
    """
    radius = config.attr(type=float, required=False, default=1.0)
    """
    The radius of the sphere.
    """

    def find_mbb(self):
        # Find the minimum bounding box, to avoid computing it every time
        mbb_min = np.array([-self.radius, -self.radius, -self.radius]) + self.origin
        mbb_max = np.array([self.radius, self.radius, self.radius]) + self.origin
        return mbb_min, mbb_max

    def get_volume(self):
        return np.pi * 4.0 / 3.0 * np.power(self.radius, 3)

    def translate(self, t_vector: np.ndarray[float]):
        self.origin += t_vector
        self.mbb_min += t_vector
        self.mbb_max += t_vector

    def rotate(self, r_versor: np.ndarray[float], angle: float):  # pragma: no cover
        # It's a sphere, it's invariant under rotation!
        pass

    def surface_function(self, theta, phi):
        """
        Convert polar coordinates into their 3D location on the sphere surface.

        :param float|numpy.ndarray[float] theta: first polar coordinate in [0; 2*np.pi]
        :param float|numpy.ndarray[float] phi: second polar coordinate in [0; np.pi]
        :return: surface coordinates
        :rtype: float|numpy.ndarray[float]
        """
        return np.array(
            [
                self.radius * np.cos(theta) * np.sin(phi),
                self.radius * np.sin(theta) * np.sin(phi),
                self.radius * np.cos(phi),
            ]
        )

    def generate_point_cloud(self, npoints: int):
        # Generate a sphere centered at the origin.
        cloud = uniform_surface_sampling(npoints, self.surface_function)
        cloud = cloud.T * np.random.rand(npoints, 3)  # sample within the shape

        cloud = cloud + self.origin

        return cloud

    def check_inside(self, points: np.ndarray[float]):
        # Translate the points, bringing the origin to the center of the sphere,
        # then check the inequality defining the sphere
        pts_centered = points - self.origin
        lhs = np.linalg.norm(pts_centered, axis=1)
        inside_points = lhs < self.radius + self.epsilon
        return inside_points

    def wireframe_points(self, nb_points_1=30, nb_points_2=30):
        x, y, z = uniform_surface_wireframe(
            nb_points_1, nb_points_2, self.surface_function
        )
        return translate_3d_mesh_by_vec(x, y, z, self.origin)


@config.node
class Cuboid(GeometricShape, classmap_entry="cuboid"):
    """
    A rectangular parallelepiped, described in cartesian coordinates.
    """

    origin = config.attr(
        type=types.ndarray(shape=(3,), dtype=float), required=True, hint=[0.0, 0.0, 0.0]
    )
    """
    The coordinates of the center of the barycenter of the bottom rectangle.
    """
    top_center = config.attr(
        type=types.ndarray(shape=(3,), dtype=float), required=True, hint=[0.0, 1.0, 0.0]
    )
    """
    The coordinates of the center of the barycenter of the top rectangle.
    """
    side_length_1 = config.attr(type=float, required=False, default=1.0)
    """
    Length of one side of the base rectangle.
    """
    side_length_2 = config.attr(type=float, required=False, default=1.0)
    """
    Length of the other side of the base rectangle.
    """

    def find_mbb(self):
        # Extrema of the cuboid centered at the origin
        extrema = [
            np.array([-self.side_length_1 / 2.0, -self.side_length_2 / 2.0, 0.0]),
            np.array([self.side_length_1 / 2.0, self.side_length_2 / 2.0, 0.0]),
            np.array([-self.side_length_1 / 2.0, self.side_length_2 / 2.0, 0.0]),
            np.array([self.side_length_1 / 2.0, -self.side_length_2 / 2.0, 0.0]),
            np.array(
                [
                    self.side_length_1 / 2.0 + self.top_center[0],
                    self.side_length_2 / 2.0 + self.top_center[1],
                    self.top_center[2],
                ]
            ),
            np.array(
                [
                    -self.side_length_1 / 2.0 + self.top_center[0],
                    self.side_length_2 / 2.0 + self.top_center[1],
                    self.top_center[2],
                ]
            ),
            np.array(
                [
                    -self.side_length_1 / 2.0 + self.top_center[0],
                    -self.side_length_2 / 2.0 + self.top_center[1],
                    self.top_center[2],
                ]
            ),
            np.array(
                [
                    self.side_length_1 / 2.0 + self.top_center[0],
                    -self.side_length_2 / 2.0 + self.top_center[1],
                    self.top_center[2],
                ]
            ),
        ]

        # Rotate the cuboid
        return _get_extrema_after_rot(extrema, self.origin, self.top_center)

    def get_volume(self):
        h = np.linalg.norm(self.top_center - self.origin)
        return h * self.side_length_1 * self.side_length_2

    def translate(self, t_vector: np.ndarray[float]):
        self.origin += t_vector
        self.top_center += t_vector
        self.mbb_min += t_vector
        self.mbb_max += t_vector

    def rotate(self, r_versor: np.ndarray[float], angle: float):
        rot = R.from_rotvec(r_versor * angle)
        self.top_center = rot.apply(self.top_center)

    def generate_point_cloud(self, npoints: int):
        # Generate a unit cuboid whose base rectangle has the barycenter in the origin
        rand = np.random.rand(npoints, 3)
        rand[:, 0] = rand[:, 0] - 0.5
        rand[:, 1] = rand[:, 1] - 0.5

        # Scale the sides of the cuboid
        height = np.linalg.norm(self.top_center - self.origin)
        rand[:, 0] = rand[:, 0] * self.side_length_1 / 2.0
        rand[:, 1] = rand[:, 1] * self.side_length_2 / 2.0
        rand[:, 2] = rand[:, 2] * height

        # Rotate the cuboid
        rot = _get_rotation_vector((self.top_center - self.origin) / height)
        cloud = rot.apply(rand)

        # Translate the cuboid
        cloud = cloud + self.origin
        return cloud

    def check_inside(self, points: np.ndarray[float]):
        # Translate back to origin
        pts = points - self.origin

        # Rotate back to xyz
        height = np.linalg.norm(self.top_center - self.origin)
        rot = _get_rotation_vector(
            (self.top_center - self.origin) / height, positive_angle=False
        )
        rot_pts = rot.apply(pts)

        # Check for intersections
        inside_points = (
            (rot_pts[:, 2] < height)
            & (rot_pts[:, 2] > 0.0)
            & (rot_pts[:, 0] < self.side_length_1)
            & (rot_pts[:, 0] > -self.side_length_1)
            & (rot_pts[:, 1] < self.side_length_2)
            & (rot_pts[:, 1] > -self.side_length_2)
        )
        return inside_points

    def wireframe_points(self, **kwargs):
        a = self.side_length_1 / 2.0
        b = self.side_length_2 / 2.0
        c = np.linalg.norm(self.top_center - self.origin)

        x = np.array(
            [
                [-a, a, a, -a],  # x coordinate of points in bottom surface
                [-a, a, a, -a],  # x coordinate of points in upper surface
                [-a, a, -a, a],  # x coordinate of points in outside surface
                [-a, a, -a, a],
            ]
        )  # x coordinate of points in inside surface
        y = np.array(
            [
                [-b, -b, b, b],  # y coordinate of points in bottom surface
                [-b, -b, b, b],  # y coordinate of points in upper surface
                [-b, -b, -b, -b],  # y coordinate of points in outside surface
                [b, b, b, b],
            ]
        )  # y coordinate of points in inside surface
        z = np.array(
            [
                [0.0, 0.0, 0.0, 0.0],  # z coordinate of points in bottom surface
                [c, c, c, c],  # z coordinate of points in upper surface
                [0.0, 0.0, c, c],  # z coordinate of points in outside surface
                [0.0, 0.0, c, c],
            ]
        )  # z coordinate of points in inside surface

        # Rotate the cuboid
        hv = (self.top_center - self.origin) / c
        return _rotate_by_coord(x, y, z, hv, self.origin)


@config.node
class Parallelepiped(GeometricShape, classmap_entry="parallelepiped"):
    """
    A generic parallelepiped, described by the vectors (following the right-hand
    orientation) of the sides in cartesian coordinates.
    """

    origin = config.attr(
        type=types.ndarray(shape=(3,), dtype=float), required=True, hint=[0.0, 0.0, 0.0]
    )
    """
    The coordinates of the left-bottom edge.
    """
    side_vector_1 = config.attr(
        type=types.ndarray(shape=(3,), dtype=float), required=True, hint=[1.0, 0.0, 0.0]
    )
    """
    The first vector identifying the parallelepiped (using the right-hand orientation:
    the thumb).
    """
    side_vector_2 = config.attr(
        type=types.ndarray(shape=(3,), dtype=float), required=True, hint=[0.0, 1.0, 0.0]
    )
    """
    The second vector identifying the parallelepiped (using the right-hand orientation:
    the index).
    """
    side_vector_3 = config.attr(
        type=types.ndarray(shape=(3,), dtype=float), required=True, hint=[0.0, 0.0, 1.0]
    )
    """
    The third vector identifying the parallelepiped (using the right-hand orientation:
    the middle finger).
    """

    def find_mbb(self):
        extrema = np.vstack(
            [
                np.array([0.0, 0.0, 0.0]),
                self.side_vector_1 + self.side_vector_2 + self.side_vector_3,
            ]
        )
        maxima = np.max(extrema, axis=0) + self.origin
        minima = np.min(extrema, axis=0) + self.origin
        return minima, maxima

    def get_volume(self):
        vol = np.dot(self.side_vector_3, np.cross(self.side_vector_1, self.side_vector_2))
        return vol

    def translate(self, t_vector: np.ndarray[float]):
        self.origin += t_vector
        self.mbb_min += t_vector
        self.mbb_max += t_vector

    def rotate(self, r_versor: np.ndarray[float], angle: float):
        rot = R.from_rotvec(r_versor * angle)
        # self.center = rot.apply(self.center)
        self.side_vector_1 = rot.apply(self.side_vector_1)
        self.side_vector_2 = rot.apply(self.side_vector_2)
        self.side_vector_3 = rot.apply(self.side_vector_3)

    def generate_point_cloud(self, npoints: int):
        # Generate a linear combination of points in the volume
        cloud = np.full((npoints, 3), 0, dtype=float)
        rand = np.random.rand(npoints, 3)
        for i in range(npoints):
            cloud[i] = (
                rand[i, 0] * self.side_vector_1
                + rand[i, 1] * self.side_vector_2
                + rand[i, 2] * self.side_vector_3
            )
        cloud += self.origin
        return cloud

    def check_inside(self, points: np.ndarray[float]):
        # Translate back to origin
        pts = points - self.origin

        # Rotate back to xyz
        height = np.linalg.norm(self.side_vector_3)
        rot = _get_rotation_vector(hv=self.side_vector_3 / height, positive_angle=True)
        rot_pts = rot.apply(pts)

        # Compute the Fourier components wrt to the vectors identifying the parallelepiped

        v1_norm = np.linalg.norm(self.side_vector_1)
        comp1 = rot_pts.dot(self.side_vector_1) / v1_norm
        v2_norm = np.linalg.norm(self.side_vector_2)
        comp2 = rot_pts.dot(self.side_vector_2) / v2_norm
        v3_norm = np.linalg.norm(self.side_vector_3)
        comp3 = rot_pts.dot(self.side_vector_3) / v3_norm

        # The points are inside the parallelepiped if and only if all the Fourier
        # components are between 0 and the norm of sides of the parallelepiped
        inside_points = (
            (comp1 > 0.0)
            & (comp1 < v1_norm)
            & (comp2 > 0.0)
            & (comp2 < v2_norm)
            & (comp3 > 0.0)
            & (comp3 < v3_norm)
        )
        return inside_points

    def wireframe_points(self, **kwargs):
        va = self.side_vector_1
        vb = self.side_vector_2
        vc = self.side_vector_3

        a = va
        b = va + vb
        c = vb
        d = np.array([0.0, 0.0, 0.0])
        e = va + vc
        f = va + vb + vc
        g = vb + vc
        h = vc

        x = np.array(
            [
                [a[0], b[0], c[0], d[0]],
                [e[0], f[0], g[0], h[0]],
                [a[0], b[0], f[0], e[0]],
                [d[0], c[0], g[0], h[0]],
            ]
        )
        y = np.array(
            [
                [a[1], b[1], c[1], d[1]],
                [e[1], f[1], g[1], h[1]],
                [a[1], b[1], f[1], e[1]],
                [d[1], c[1], g[1], h[1]],
            ]
        )
        z = np.array(
            [
                [a[2], b[2], c[2], d[2]],
                [e[2], f[2], g[2], h[2]],
                [a[2], b[2], f[2], e[2]],
                [d[2], c[2], g[2], h[2]],
            ]
        )

        return x + self.origin[0], y + self.origin[1], z + self.origin[2]


__all__ = [
    "GeometricShape",
    "ShapesComposition",
    "Ellipsoid",
    "Cone",
    "Cylinder",
    "Sphere",
    "Cuboid",
    "Parallelepiped",
    "inside_mbox",
]



import numpy as np

from ... import config
from ...config import types
from ...trees import BoxTree
from .. import ConnectionStrategy
from ..strategy import Hemitype
from .geometric_shapes import ShapesComposition


@config.node
class ShapeHemitype(Hemitype):
    """
    Class representing a population of cells to connect with a ConnectionStrategy.

    These cells' morphology is implemented as a ShapesComposition.
    """

    shapes_composition = config.attr(type=ShapesComposition, required=True)
    """
    Composite shape representing the Hemitype.
    """

    def get_mbb(self, chunks, chunk_dimension):
        """
        Get the list of minimal bounding box containing all cells in the `ShapeHemitype`.

        :param chunks: List of chunks containing the cell types (see
            bsb.connectivity.strategy.Hemitype.get_all_chunks)
        :type chunks: list[bsb.storage._chunks.Chunk]
        :param chunk_dimension: Size of a chunk
        :type chunk_dimension: float
        :return: List of bounding boxes in the form [min_x, min_y, min_z, max_x, max_y,
            max_z] for each chunk containing cells.
        :rtype: list[numpy.ndarray[float, float, float, float, float, float]]
        """
        return [
            np.concatenate(
                [
                    self.shapes_composition.get_mbb_min()
                    + np.array(idx_chunk) * chunk_dimension,
                    self.shapes_composition.get_mbb_max()
                    + np.array(idx_chunk) * chunk_dimension,
                ]
            )
            for idx_chunk in chunks
        ]

    def _get_shape_boxtree(self, chunks):
        mbbs = self.get_mbb(chunks, chunks[0].dimensions)
        return BoxTree(mbbs)


@config.node
class ShapeToShapeIntersection(ConnectionStrategy):
    presynaptic = config.attr(type=ShapeHemitype, required=True)
    postsynaptic = config.attr(type=ShapeHemitype, required=True)
    affinity = config.attr(type=types.fraction(), required=True, hint=0.1)
    """
    Ratio of apositions to keep over the total number of contact points.
    """
    pruning_ratio = config.attr(type=types.fraction(), required=True, hint=0.1)
    """
    Ratio of conections to keep over the total number of apositions.
    """

    def get_region_of_interest(self, chunk):
        # Filter postsyn chunks that overlap the presyn chunk.
        post_chunks = self.postsynaptic.get_all_chunks()
        tree = self.postsynaptic._get_shape_boxtree(
            post_chunks,
        )
        pre_mbb = self.presynaptic.get_mbb(
            self.presynaptic.get_all_chunks(), chunk.dimensions
        )
        return [post_chunks[i] for i in tree.query(pre_mbb, unique=True)]

    def connect(self, pre, post):
        for pre_ps in pre.placement:
            for post_ps in post.placement:
                self._connect_type(pre_ps.cell_type, pre_ps, post_ps.cell_type, post_ps)

    def _connect_type(self, pre_ct, pre_ps, post_ct, post_ps):
        pre_pos = pre_ps.load_positions()
        post_pos = post_ps.load_positions()

        pre_shapes_cache = self.presynaptic.shapes_composition.__copy__()
        post_shapes_cache = self.postsynaptic.shapes_composition.__copy__()

        to_connect_pre = np.empty([0, 3], dtype=int)
        to_connect_post = np.empty([0, 3], dtype=int)

        for pre_id, pre_coord in enumerate(pre_pos):
            # Generate pre point cloud
            pre_shapes_cache.translate(pre_coord)
            pre_point_cloud = pre_shapes_cache.generate_point_cloud()

            def find_mbb(coords):
                maxima = np.max(coords, axis=0)
                minima = np.min(coords, axis=0)
                return minima, maxima

            def BoxesOverlap(box1min, box1max, box2min, box2max):
                return np.all((box1max >= box2min) & (box2max >= box1min))

            pre_mbb_min = pre_shapes_cache.get_mbb_min()
            pre_mbb_max = pre_shapes_cache.get_mbb_max()

            points_per_cloud = int(len(pre_point_cloud) * self.affinity)
            tmp_pre_selection = np.full(
                [len(post_pos) * int(points_per_cloud), 3], -1, dtype=int
            )
            tmp_post_selection = np.full(
                [len(post_pos) * int(points_per_cloud), 3], -1, dtype=int
            )
            ptr = 0
            for post_id, post_coord in enumerate(post_pos):
                post_shapes_cache.translate(post_coord)
                post_mbb_min = post_shapes_cache.get_mbb_min()
                post_mbb_max = post_shapes_cache.get_mbb_max()
                boxes_overlap = BoxesOverlap(
                    post_mbb_min, post_mbb_max, pre_mbb_min, pre_mbb_max
                )
                if boxes_overlap:
                    # Compare pre and post mbbs
                    inside_mbbox = post_shapes_cache.inside_mbox(pre_point_cloud)
                    if np.any(inside_mbbox):
                        inside_pts = post_shapes_cache.inside_shapes(pre_point_cloud)
                        selected = pre_point_cloud[inside_pts]

                        def sizemod(q, aff):
                            ln = len(q)
                            return int(
                                np.floor(ln * aff) + (np.random.rand() < ((ln * aff) % 1))
                            )

                        selected = selected[
                            np.random.randint(
                                len(selected), size=sizemod(selected, self.affinity)
                            ),
                            :,
                        ]
                        n_synapses = len(selected)
                        if n_synapses > 0:
                            tmp_pre_selection[ptr : ptr + n_synapses, 0] = pre_id
                            tmp_post_selection[ptr : ptr + n_synapses, 0] = post_id
                            ptr += n_synapses
                post_shapes_cache.translate(-post_coord)
            if ptr > 0:
                to_connect_pre = np.vstack([to_connect_pre, tmp_pre_selection[:ptr]])
                to_connect_post = np.vstack([to_connect_post, tmp_post_selection[:ptr]])

            pre_shapes_cache.translate(-pre_coord)

        if self.pruning_ratio < 1 and len(to_connect_pre) > 0:
            ids_to_select = np.random.choice(
                len(to_connect_pre),
                int(np.floor(self.pruning_ratio * len(to_connect_pre))),
                replace=False,
            )
            to_connect_pre = to_connect_pre[ids_to_select]
            to_connect_post = to_connect_post[ids_to_select]

        self.connect_cells(pre_ps, post_ps, to_connect_pre, to_connect_post)


__all__ = ["ShapeHemitype", "ShapeToShapeIntersection"]



# ruff: noqa: F401
# This barrel file contains intentional import shortcuts
# isort: off
# Load module before others to prevent partially initialized modules
from .strategy import ConnectionStrategy

# isort: on
from .detailed import VoxelIntersection
from .general import AllToAll, Convergence, FixedIndegree, FixedOutdegree
from .geometric import (
    Cone,
    Cuboid,
    Cylinder,
    Ellipsoid,
    GeometricShape,
    MorphologyToShapeIntersection,
    Parallelepiped,
    ShapeHemitype,
    ShapesComposition,
    ShapeToMorphologyIntersection,
    ShapeToShapeIntersection,
    Sphere,
    inside_mbox,
)
from .import_ import CsvImportConnectivity



import numpy as np

from ... import config
from ...config import types
from .. import ConnectionStrategy
from .shape_shape_intersection import ShapeHemitype


def _create_geometric_conn_arrays(branches, ids, coord):
    morpho_points = 0
    for b in branches:
        morpho_points += len(b.points)
    points_ids = np.empty([morpho_points, 3], dtype=int)
    morpho_coord = np.empty([morpho_points, 3], dtype=float)
    local_ptr = 0
    for i, b in enumerate(branches):
        points_ids[local_ptr : local_ptr + len(b.points), 0] = ids
        points_ids[local_ptr : local_ptr + len(b.points), 1] = i
        points_ids[local_ptr : local_ptr + len(b.points), 2] = np.arange(len(b.points))
        tmp = b.points + coord
        morpho_coord[local_ptr : local_ptr + len(b.points), :] = tmp
        local_ptr += len(b.points)
    return points_ids, morpho_coord


@config.node
class ShapeToMorphologyIntersection(ConnectionStrategy):
    presynaptic = config.attr(type=ShapeHemitype, required=True)
    affinity = config.attr(type=types.fraction(), required=True, hint=0.1)
    """
    Ratio of apositions to keep over the total number of contact points.
    """
    pruning_ratio = config.attr(type=types.fraction(), required=True, hint=0.1)
    """
    Ratio of conections to keep over the total number of apositions.
    """

    def get_region_of_interest(self, chunk):
        lpost, upost = self.postsynaptic._get_rect_ext(tuple(chunk.dimensions))
        pre_chunks = self.presynaptic.get_all_chunks()
        tree = self.presynaptic._get_shape_boxtree(
            pre_chunks,
        )
        post_mbb = [
            np.concatenate(
                [
                    (lpost + chunk) * chunk.dimensions,
                    np.maximum((upost + chunk), (lpost + chunk) + 1) * chunk.dimensions,
                ]
            )
        ]

        return [pre_chunks[i] for i in tree.query(post_mbb, unique=True)]

    def connect(self, pre, post):
        for pre_ps in pre.placement:
            for post_ps in post.placement:
                self._connect_type(pre_ps.cell_type, pre_ps, post_ps.cell_type, post_ps)

    def _connect_type(self, pre_ct, pre_ps, post_ct, post_ps):
        pre_pos = pre_ps.load_positions()
        post_pos = post_ps.load_positions()

        pre_shapes = self.presynaptic.shapes_composition.__copy__()

        to_connect_pre = np.empty([0, 3], dtype=int)
        to_connect_post = np.empty([0, 3], dtype=int)

        morpho_set = post_ps.load_morphologies()
        post_morphos = morpho_set.iter_morphologies(cache=True, hard_cache=True)

        for post_id, (post_coord, morpho) in enumerate(
            zip(post_pos, post_morphos, strict=False)
        ):
            # Get the branches
            branches = morpho.get_branches()

            # Build ids array from the morphology
            post_points_ids, post_morpho_coord = _create_geometric_conn_arrays(
                branches, post_id, post_coord
            )

            for pre_id, pre_coord in enumerate(pre_pos):
                pre_shapes.translate(pre_coord)
                mbb_check = pre_shapes.inside_mbox(post_morpho_coord)

                if np.any(mbb_check):
                    inside_pts = pre_shapes.inside_shapes(post_morpho_coord[mbb_check])
                    # Find the morpho points inside the postsyn geometric shapes
                    if np.any(inside_pts):
                        local_selection = (post_points_ids[mbb_check])[inside_pts]
                        if self.affinity < 1.0 and len(local_selection) > 0:
                            nb_targets = np.max(
                                [1, int(np.floor(self.affinity * len(local_selection)))]
                            )
                            chosen_targets = np.random.choice(
                                local_selection.shape[0], nb_targets
                            )
                            local_selection = local_selection[chosen_targets, :]
                        selected_count = len(local_selection)
                        if selected_count > 0:
                            to_connect_post = np.vstack(
                                [to_connect_post, local_selection]
                            )
                            pre_tmp = np.full([selected_count, 3], -1, dtype=int)
                            pre_tmp[:, 0] = pre_id
                            to_connect_pre = np.vstack([to_connect_pre, pre_tmp])
                pre_shapes.translate(-pre_coord)
        if self.pruning_ratio < 1 and len(to_connect_pre) > 0:
            ids_to_select = np.random.choice(
                len(to_connect_pre),
                int(np.floor(self.pruning_ratio * len(to_connect_pre))),
                replace=False,
            )
            to_connect_pre = to_connect_pre[ids_to_select]
            to_connect_post = to_connect_post[ids_to_select]
        self.connect_cells(pre_ps, post_ps, to_connect_pre, to_connect_post)


__all__ = ["ShapeToMorphologyIntersection"]



"""
This module imports all supported storage engines, objects that read and write data, which
are present as subfolders of the `engine` folder, and provides them transparently to the
user, as a part of the :class:`Storage <.storage.Storage>` factory class. The module scans
the :mod:`.storage.interfaces` module for any class that inherits from :class:`Interface
<.storage.interfaces.Interface>`  to collect all Feature Interfaces and then scans the
``storage.engines.*`` submodules for any class that provides an implementation of those
features.

These features, because they all follow the same interface can then be passed on to
consumers and can be used independent of the underlying storage engine, which is the end
goal of this module.
"""

import functools
import typing
from inspect import isclass

from .. import plugins
from ..exceptions import UnknownStorageEngineError
from ..services.mpi import MPIService

if typing.TYPE_CHECKING:
    from .interfaces import (
        ConnectivitySet,
        FileStore,
        MorphologyRepository,
        PlacementSet,
    )


@functools.cache
def get_storage_interfaces():
    from . import interfaces

    # Collect all classes that are a subclass of Interface except Interface itself and
    # store them in a {class_name: class_object} dictionary
    return {
        interface.__name__: interface
        for interface in interfaces.__dict__.values()
        if isclass(interface)
        and issubclass(interface, interfaces.Interface)
        and interface is not interfaces.Interface
    }


@functools.cache
def discover_engines():
    """
    Get a dictionary of all available storage engines.
    """
    return plugins.discover("storage.engines")


@functools.cache
def get_engine_support(engine_name):
    try:
        engine_module = discover_engines()[engine_name]
    except KeyError:
        raise UnknownStorageEngineError(
            f"Unknown storage engine '{engine_name}'"
        ) from None
    engine_support = {
        interface_name: NotSupported(interface_name)
        for interface_name in get_storage_interfaces()
    }
    engine_support["StorageNode"] = engine_module.StorageNode
    # Search for interface support
    for interface_name, interface in get_storage_interfaces().items():
        for module_item in engine_module.__dict__.values():
            # Look through module items for child class of interface
            if (
                isclass(module_item)
                and module_item is not interface
                and issubclass(module_item, interface)
            ):
                engine_support[interface_name] = module_item
                break

    return engine_support


@functools.cache
def get_engines():
    return {name: get_engine_support(name)["Engine"] for name in discover_engines()}


def create_engine(name, root, comm):
    """
    Create an engine from the engine's Engine interface.

    :param str name: The name of the engine to create.
    :param object root: An object that uniquely describes the storage, such as a filename
        or path. The value to be provided depends on the engine. For the hdf5 engine the
        filename has to be provided.
    :param bsb.services.mpi.MPIService comm: MPI communicator that shares control over the
        Engine interface.
    """
    return get_engine_support(name)["Engine"](root, comm)


class NotSupported:
    """
    Utility class that throws a ``NotSupported`` error when it is used.

    This is the default "implementation" of every storage feature that isn't provided by
    an engine.
    """

    _iface_engine_key = None

    def __init__(self, operation):
        self.operation = operation

    def _unsupported_err(self):
        # Throw an error detailing the lack of support of our engine for our feature.
        raise NotImplementedError(
            f"The storage engine does not support the {self.operation} feature"
        )

    def __call__(self, *args, **kwargs):
        self._unsupported_err()

    def __getattr__(self, attr):
        self._unsupported_err()


class Storage:
    """
    Factory class that produces all the features and shims the functionality of the
    underlying engine.
    """

    _PlacementSet: type["PlacementSet"]
    _ConnectivitySet: type["ConnectivitySet"]
    _MorphologyRepository: type["MorphologyRepository"]
    _FileStore: type["FileStore"]

    def __init__(self, engine, root, comm=None, main=0, missing_ok=True):
        """
        Create a Storage provider based on a specific `engine` uniquely identified by the
        root object.

        :param engine: The name of the storage engine.
        :type engine: str
        :param root: An object that uniquely describes the storage, such as a filename or
            path. The value to be provided depends on the engine. For the hdf5 engine the
            filename has to be provided.
        :type root: object
        :param comm: MPI communicator that shares control over this Storage.
        :type comm: mpi4py.MPI.Comm
        :param main: Rank of the MPI process that executes single-node tasks.
        """
        self._comm = MPIService(comm)
        self._engine = create_engine(engine, root, self._comm)
        self._features = [
            fname for fname, supported in view_support()[engine].items() if supported
        ]
        self._engine._format = engine
        self._main = main

        # Load the engine's interface onto the object, this allows the end user to create
        # features, but it is not advised. Usually the Storage object
        # itself provides factory methods that should be used instead.
        for name, interface in get_engine_support(engine).items():
            if name == "StorageNode":
                continue
            self.__dict__["_" + name] = interface
            # Interfaces can define an autobinding key so that singletons are available
            # on the engine under that key.
            key = interface._iface_engine_key
            if key is not None:
                if self.supports(name):
                    self._engine.__dict__[key] = interface(self._engine)
                else:
                    self._engine.__dict__[key] = NotSupported(self._engine.format, name)
        # The storage should be created at the root as soon as we initialize because
        # features might immediatly require the basic structure to be present.
        self._preexisted = self.exists()
        if not (missing_ok or self._preexisted):
            raise FileNotFoundError(f"`{engine}` storage at '{root}' does not exist.")
        if not self._preexisted:
            self.create()

    def __eq__(self, other):
        return self._engine == getattr(other, "_engine", None)

    @property
    def preexisted(self):
        return self._preexisted

    def is_main_process(self):
        return self._comm.get_rank() == self._main

    @property
    def morphologies(self):
        return self._engine.morphologies

    @property
    def files(self):
        return self._engine.files

    @property
    def root(self):
        return self._engine.root

    @property
    def root_slug(self):
        return self._engine.root_slug

    @property
    def format(self):
        return self._engine._format

    def exists(self):
        """
        Check whether the storage exists at the root.
        """
        return self._engine.exists()

    def create(self):
        """
        Create the minimal requirements at the root for other features to function and for
        the existence check to pass.
        """
        return self._engine.create()

    def copy(self, new_root):
        """
        Move the storage to a new root.
        """
        self._engine.copy(new_root)

    def move(self, new_root):
        """
        Move the storage to a new root.
        """
        self._engine.move(new_root)

    def remove(self):
        """
        Remove the storage and all data contained within.

        This is an irreversible destructive action!
        """
        self._engine.remove()

    def load(self):
        """
        Load a scaffold from the storage.

        :returns: :class:`Scaffold <.core.Scaffold>`
        """
        from ..core import Scaffold

        return Scaffold(storage=self, comm=self._comm.get_communicator())

    def load_active_config(self):
        """
        Load the configuration object from the storage.

        :returns: :class:`Configuration <.config.Configuration>`
        """
        return self._engine.files.load_active_config()

    def store_active_config(self, config):
        """
        Store a configuration object in the storage.
        """
        return self._engine.files.store_active_config(config)

    def supports(self, feature):
        return feature in self._features

    def assert_support(self, feature):
        if not self.supports(feature):
            raise NotImplementedError(
                f"The '{self._engine._format}' engine lacks support "
                f"for the '{feature}' feature."
            )

    def get_placement_set(self, type, chunks=None, labels=None, morphology_labels=None):
        """
        Return a PlacementSet for the given type.

        :param type: Specific cell type.
        :type type: ~bsb.cell_types.CellType
        :param chunks: Optionally load a specific list of chunks.
        :type chunks: list[tuple[float, float, float]]
        :param labels: Labels to filter the placement set by.
        :type labels: list[str]
        :returns: ~bsb.storage.interfaces.PlacementSet
        """
        ps = self._PlacementSet(self._engine, type)
        if chunks is not None:
            ps.set_chunk_filter(chunks)
        ps.set_label_filter(labels)
        ps.set_morphology_label_filter(morphology_labels)
        return ps

    def require_placement_set(self, cell_type):
        """
        Get a placement set.

        :param cell_type: Connection cell_type
        :type cell_type: ~bsb.cell_types.CellType
        :returns: ~bsb.storage.interfaces.PlacementSet
        """
        return self._PlacementSet.require(self._engine, cell_type)

    def get_connectivity_set(self, tag):
        """
        Get a connection set.

        :param tag: Connection tag
        :type tag: str
        :returns: ~bsb.storage.interfaces.ConnectivitySet
        """
        return self._ConnectivitySet(self._engine, tag)

    def require_connectivity_set(self, tag, pre=None, post=None):
        """
        Get a connection set.

        :param tag: Connection tag
        :type tag: str
        :returns: ~bsb.storage.interfaces.ConnectivitySet
        """
        return self._ConnectivitySet.require(self._engine, tag, pre, post)

    def get_connectivity_sets(self):
        """
        Return a ConnectivitySet for the given type.

        :param type: Specific cell type.
        :type type: ~bsb.cell_types.CellType
        :returns: ~bsb.storage.interfaces.ConnectivitySet
        """
        return [
            self._ConnectivitySet(self._engine, tag)
            for tag in self._ConnectivitySet.get_tags(self._engine)
        ]

    def init(self, scaffold):
        """
        Initialize the storage to be ready for use by the specified scaffold.
        """
        self.store_active_config(scaffold.configuration)
        if self.supports("PlacementSet"):
            self.init_placement(scaffold)

    def init_placement(self, scaffold):
        for cell_type in scaffold.get_cell_types():
            self.require_placement_set(cell_type)

    def renew(self, scaffold):
        """
        Remove and recreate an empty storage container for a scaffold.
        """
        self.remove()
        self.create()
        self.init(scaffold)

    def clear_placement(self, scaffold=None):
        self._engine.clear_placement()
        if scaffold is not None:
            self.init_placement(scaffold)

    def clear_connectivity(self):
        self._engine.clear_connectivity()

    def read_only(self):
        return self._engine.read_only()

    def get_chunk_stats(self):
        return self._engine.get_chunk_stats()


def open_storage(root, comm=None):
    """
    Load a Storage object from its root.

    :param root: Root (usually path) pointing to the storage object.
    :param mpi4py.MPI.Comm comm: MPI communicator that shares control
      over the Storage.
    :returns: A network scaffold
    :rtype: :class:`Storage`
    """
    engines = get_engines()
    for name, engine in engines.items():
        if engine.peek_exists(root) and engine.recognizes(root, comm):
            return Storage(name, root, comm, missing_ok=False)
    else:
        for engine in engines.values():
            if engine.peek_exists(root):
                raise OSError(
                    f"Storage `{root}` not recognized as any installed format: "
                    + ", ".join(f"'{n}'" for n in engines)
                )
        else:
            raise FileNotFoundError(f"Storage `{root}` does not exist.")


def get_engine_node(engine_name):
    try:
        return get_engine_support(engine_name)["StorageNode"]
    except KeyError:
        raise RuntimeError(
            f"Broken storage engine plugin '{engine_name}' is missing a StorageNode."
        ) from None


def view_support(engine=None):
    """
    Return which storage engines support which features.
    """
    if engine is None:
        return {
            # Loop over all engines
            engine_name: {
                # Loop over all features, check whether they're supported
                feature_name: not isinstance(feature, NotSupported)
                for feature_name, feature in get_engine_support(engine_name).items()
            }
            for engine_name in discover_engines()
        }
    else:
        # Loop over all features for the specific engine
        return {
            feature_name: not isinstance(feature, NotSupported)
            for feature_name, feature in get_engine_support(engine).items()
        }


__all__ = [
    "NotSupported",
    "Storage",
    "create_engine",
    "discover_engines",
    "get_engine_node",
    "get_engines",
    "open_storage",
    "view_support",
]



import abc

import numpy as np

from . import config
from .config import refs
from .exceptions import MorphologyDataError, MorphologyError
from .reporting import report


@config.dynamic(attr_name="strategy")
class AfterPlacementHook(abc.ABC):
    name: str = config.attr(key=True)

    def queue(self, pool):
        def static_function(scaffold, name):
            return scaffold.after_placement[name].postprocess()

        pool.queue(static_function, (self.name,), submitter=self)

    @abc.abstractmethod
    def postprocess(self):
        pass


@config.dynamic(attr_name="strategy")
class AfterConnectivityHook(abc.ABC):
    name: str = config.attr(key=True)

    def queue(self, pool):
        def static_function(scaffold, name):
            return scaffold.after_connectivity[name].postprocess()

        pool.queue(static_function, (self.name,), submitter=self)

    @abc.abstractmethod
    def postprocess(self):
        pass


class SpoofDetails(AfterConnectivityHook):
    """
    Create fake morphological intersections between already connected non-detailed
    connection types.
    """

    casts = {"presynaptic": str, "postsynaptic": str}

    def postprocess(self):
        # Check which connection types exist between the pre- and postsynaptic types.
        connection_results = self.scaffold.get_connection_cache_by_cell_type(
            presynaptic=self.presynaptic, postsynaptic=self.postsynaptic
        )
        # Iterate over each involved connectivity matrix
        for connection_result in connection_results:
            connection_type = connection_result[0]
            for connectivity_matrix in connection_result[1:]:
                # Spoof details (morphology & section intersection) between the
                # non-detailed connections in the connectivity matrix.
                self.spoof_connections(connection_type, connectivity_matrix)

    def spoof_connections(self, connection_type, connectivity_matrix):
        from_type = connection_type.presynaptic.type
        to_type = connection_type.postsynaptic.type
        from_entity = False
        to_entity = False
        # Check whether any of the types are relays or entities.
        if from_type.entity:
            from_entity = True
            if to_type.entity:
                raise MorphologyError(
                    "Can't spoof detailed connections between 2 entity cell types."
                )
        elif to_type.entity:
            to_entity = True
        # If they're not relays or entities, load their morphologies
        if not from_entity:
            from_morphologies = from_type.list_all_morphologies()
            if len(from_morphologies) == 0:
                raise MorphologyDataError(
                    "Can't spoof detailed connection without morphologies for "
                    f"'{from_type.name}'"
                )
        if not to_entity:
            to_morphologies = to_type.list_all_morphologies()
            if len(to_morphologies) == 0:
                raise MorphologyDataError(
                    "Can't spoof detailed connection without morphologies for "
                    f"'{to_type.name}'"
                )
        # If they are entities or relays,
        # steal the first morphology of the other cell type.
        # Under no circumstances should entities or relays be represented as actual
        # morphologies, so this should not matter: the data just needs to be spoofed for
        # other parts of the scaffold to function.
        if from_entity:
            from_morphologies = [to_morphologies[0]]
        if to_entity:
            to_morphologies = [from_morphologies[0]]

        # Use only the first morphology for spoofing.
        # At a later point which morphology belongs to which cell should be decided
        # as a property of the cell and not the connection.
        # At that point we can spoof the same morphologies to the opposing relay type.
        #
        # The left column will be the first from_morphology (0) and the right column
        # will be the first to_morphology (1)
        _from = np.zeros(len(connectivity_matrix))
        _to = np.ones(len(connectivity_matrix))
        morphologies = np.column_stack((_from, _to))
        # Generate the map
        morpho_map = [from_morphologies[0], to_morphologies[0]]
        from_m = self.scaffold.morphology_repository.load(from_morphologies[0])
        to_m = self.scaffold.morphology_repository.load(to_morphologies[0])
        # Select random axons and dendrites to connect
        axons = np.array(from_m.get_compartment_submask(["axon"]))
        dendrites = np.array(to_m.get_compartment_submask(["dendrites"]))
        compartments = np.column_stack(
            (
                axons[np.random.randint(0, len(axons), len(connectivity_matrix))],
                dendrites[np.random.randint(0, len(dendrites), len(connectivity_matrix))],
            )
        )
        # Erase previous connection data so that `.connect_cells` can overwrite it.
        self.scaffold.cell_connections_by_tag[connection_type.name] = np.empty((0, 2))
        # Write the new spoofed connection data
        self.scaffold.connect_cells(
            connection_type,
            connectivity_matrix,
            morphologies=morphologies,
            compartments=compartments,
            morpho_map=morpho_map,
        )
        report(
            f"Spoofed details of {len(connectivity_matrix)} connections between "
            f"{connection_type.presynaptic.type.name} and "
            f"{connection_type.postsynaptic.type.name}",
            level=2,
        )


@config.node
class Relay(AfterConnectivityHook):
    """
    Replaces connections on a cell with the relayed connections to the connection targets
    of that cell.

    Not implemented yet.
    """

    cell_types = config.reflist(refs.cell_type_ref)

    def postprocess(self):
        pass


class BidirectionalContact(AfterConnectivityHook):
    # Replicates all contacts (connections and compartments) to have bidirection in gaps
    def postprocess(self):
        for type in self.types:
            self.scaffold.cell_connections_by_tag[type] = self._invert_append(
                self.scaffold.cell_connections_by_tag[type]
            )
            self.scaffold.connection_compartments[type] = self._invert_append(
                self.scaffold.connection_compartments[type]
            )
            self.scaffold.connection_morphologies[type] = self._invert_append(
                self.scaffold.connection_morphologies[type]
            )

    def _invert_append(self, old):
        return np.concatenate((old, np.stack((old[:, 1], old[:, 0]), axis=1)), axis=0)


__all__ = [
    "BidirectionalContact",
    "AfterPlacementHook",
    "AfterConnectivityHook",
    "Relay",
    "SpoofDetails",
]



import abc
import concurrent
import contextlib
import re
import tempfile
import typing
import urllib
import warnings
from concurrent.futures import ThreadPoolExecutor

import requests

from .. import config
from ..config import types
from ..config._attrs import cfglist
from ..exceptions import MissingMorphologyError, SelectorError
from . import Morphology

if typing.TYPE_CHECKING:
    from ..core import Scaffold


@config.dynamic(
    attr_name="select",
    auto_classmap=True,
    required=False,
    default="by_name",
)
class MorphologySelector(abc.ABC):
    scaffold: "Scaffold"

    @abc.abstractmethod
    def validate(self, all_morphos):
        pass

    @abc.abstractmethod
    def pick(self, morphology):
        pass


@config.node
class NameSelector(MorphologySelector, classmap_entry="by_name"):
    names: cfglist[str] = config.list(type=str, required=types.shortform())

    def __init__(self, name=None, /, **kwargs):
        if name is not None:
            self.names = [name]

    def __inv__(self):
        if self._config_pos_init:
            return self.names[0]
        return self.__tree__()

    def _cache_patterns(self):
        self._pnames = {n: n.replace("*", r".*").replace("|", "\\|") for n in self.names}
        self._patterns = {n: re.compile(f"^{pat}$") for n, pat in self._pnames.items()}
        self._empty = not self.names
        self._match = re.compile(f"^({'|'.join(self._pnames.values())})$")

    def validate(self, all_morphos):
        self._cache_patterns()
        repo_names = {m.get_meta()["name"] for m in all_morphos}
        missing = [
            n
            for n, pat in self._patterns.items()
            if not any(pat.match(rn) for rn in repo_names)
        ]
        if missing:
            err = "Morphology repository misses the following morphologies"
            if self._config_parent is not None:
                node = self._config_parent._config_parent
                err += f" required by {node.get_node_name()}"
            err += f": {', '.join(missing)}"
            raise MissingMorphologyError(err)

    def pick(self, morphology):
        self._cache_patterns()
        return (
            not self._empty
            and self._match.match(morphology.get_meta()["name"]) is not None
        )


@config.node
class NeuroMorphoSelector(NameSelector, classmap_entry="from_neuromorpho"):
    _url = "http://cng.gmu.edu:8080/neuroMorpho/"  # "https://neuromorpho.org/"
    _meta = "api/neuron/select?q=neuron_name:"
    _files = "dableFiles/"

    def __boot__(self):
        if self.scaffold.is_main_process():
            try:
                morphos = self._scrape_nm(self.names)
            except:
                self.scaffold._comm.barrier()
                raise
            for name, morpho in morphos.items():
                self.scaffold.morphologies.save(name, morpho, overwrite=True)
        self.scaffold._comm.barrier()

    @classmethod
    def _swc_url(cls, archive, name):
        return (
            f"{cls._url}{cls._files}{urllib.parse.quote(archive.lower())}/"
            f"CNG%20version/{name}.CNG.swc"
        )

    @classmethod
    def _scrape_nm(cls, names):
        # Weak DH key on neuromorpho.org
        # https://stackoverflow.com/questions/38015537/python-requests-exceptions-sslerror-dh-key-too-small
        requests.packages.urllib3.util.ssl_.DEFAULT_CIPHERS += ":HIGH:!DH:!aNULL"
        # Pass if no pyopenssl support used / needed / available
        with contextlib.suppress(AttributeError):
            requests.packages.urllib3.contrib.pyopenssl.util.ssl_.DEFAULT_CIPHERS += (
                ":HIGH:!DH:!aNULL"
            )
        with warnings.catch_warnings():
            warnings.simplefilter("ignore")
            with ThreadPoolExecutor() as executor:
                # Certificate issues with neuromorpho --> verify=False
                res = requests.get(cls._url + cls._meta + ",".join(names), verify=False)
                if res.status_code == 404:
                    raise SelectorError(f"'{names[0]}' is not a valid NeuroMorpho name.")
                elif res.status_code != 200:
                    raise SelectorError("NeuroMorpho API error: " + res.message)
                metas = {n: None for n in names}
                for meta in res.json()["_embedded"]["neuronResources"]:
                    del meta["_links"]
                    metas[meta["neuron_name"]] = meta
                missing = [name for name, meta in metas.items() if meta is None]
                if missing:
                    raise SelectorError(
                        ", ".join(f"'{n}'" for n in missing)
                        + " are not valid NeuroMorpho names."
                    )
                swc_urls = {n: cls._swc_url(metas[n]["archive"], n) for n in names}

                def req(n):
                    return requests.get(swc_urls[n], verify=False)

                def sub(n):
                    return executor.submit(req, n), n

                futures = dict(map(sub, names))
                morphos = {n: None for n in names}
                with tempfile.TemporaryDirectory() as tempdir:
                    for future in concurrent.futures.as_completed(futures.keys()):
                        name = futures[future]
                        path = tempdir + f"/{name}.swc"
                        with open(path, "w") as f:
                            f.write(future.result().text)
                        morphos[name] = Morphology.from_swc(path, meta=metas[name])
                missing = [name for name, m in morphos.items() if m is None]
                if missing:  # pragma: nocover
                    raise SelectorError(
                        "Downloading NeuroMorpho failed for "
                        + ", ".join(f"'{n}'" for n in missing)
                        + "."
                    )
                return morphos


__all__ = ["MorphologySelector", "NameSelector", "NeuroMorphoSelector"]



import abc

import numpy as np


class Layout:
    """
    Container class for all types of partition data.

    The layout swaps the data of the partition with temporary layout associated data, and
    tries out experimental changes to the partition data, if the layout process fails, the
    original partition data is reinstated.
    """

    def __init__(self, data, owner=None, children=None, frozen=False):
        if children is None:
            children = []
        self._data = data
        self._owner = owner
        self._children = children
        self._frozen = frozen

    @property
    def data(self):
        return self._data

    @property
    def children(self):
        return self._children

    def copy(self):
        return Layout(
            data=self._data.copy(),
            owner=self._owner,
            children=self._children,
            frozen=self._frozen,
        )

    def accept(self):
        self.swap()

    def __getattr__(self, attr):
        if attr.startswith("propose_"):
            f = getattr(self._owner, attr[8:])

            def swapped_execute(*args, **kwargs):
                self.swap()
                f(*args, **kwargs)
                self.swap()

            return swapped_execute

        super().__getattribute__(attr)

    def swap(self):
        if self._owner is not None:
            old = getattr(self._owner, "_data", None)
            self._owner._data = self._data
            self._data = old
        for child in self._children:
            child.swap()


class PartitionData(abc.ABC):
    """
    The partition data is a class that stores the description of a partition for a
    partition.

    This allows the Partition interface to define mutating operations such as translate,
    rotate, scale; for a dry-run we only have to swap out the actual data with temporary
    data, and the mutation is prevented.
    """

    @abc.abstractmethod
    def copy(self):
        pass


class RhomboidData(PartitionData):
    def __init__(self, ldc, mdc):
        # Least dominant corner
        self.ldc = np.array(ldc, dtype=float, copy=False)
        # Most dominant corner
        self.mdc = np.array(mdc, dtype=float, copy=False)

    def __repr__(self):
        return f"{self.__class__.__name__}({self.ldc}, {self.mdc})"

    def copy(self):
        """
        Copy this boundary to a new instance.
        """
        return self.__class__(self.ldc.copy(), self.mdc.copy())

    @property
    def x(self):
        return self.ldc[0]

    @x.setter
    def x(self, value):
        self.mdc[0] += value - self.ldc[0]
        self.ldc[0] = value

    @property
    def y(self):
        return self.ldc[1]

    @y.setter
    def y(self, value):
        self.mdc[1] += value - self.ldc[1]
        self.ldc[1] = value

    @property
    def z(self):
        return self.ldc[2]

    @z.setter
    def z(self, value):
        self.mdc[2] += value - self.ldc[2]
        self.ldc[2] = value

    @property
    def dimensions(self):
        return self.mdc - self.ldc

    @property
    def width(self):
        return self.mdc[0] - self.ldc[0]

    @width.setter
    def width(self, value):
        self.mdc[0] = self.ldc[0] + value

    @property
    def height(self):
        return self.mdc[2] - self.ldc[2]

    @height.setter
    def height(self, value):
        self.mdc[2] = self.ldc[2] + value

    @property
    def depth(self):
        return self.mdc[1] - self.ldc[1]

    @depth.setter
    def depth(self, value):
        self.mdc[1] = self.ldc[1] + value


def box_layout(ldc, mdc):
    return Layout(RhomboidData(ldc, mdc), frozen=True)



from __future__ import annotations

from .. import warn
from ..exceptions import ConfigurationWarning, FileImportError, ParserError


class parsed_node:
    _key = None
    """
    Key to reference the node.
    """
    _parent: parsed_node = None
    """
    Parent node.
    """

    def location(self):
        return "/" + "/".join(str(part) for part in self._location_parts([]))

    def _location_parts(self, carry):
        parent = self
        while (parent := parent._parent) is not None:
            if parent._parent is not None:
                carry.insert(0, parent._key)
        carry.append(self._key or "")
        return carry

    def __str__(self):
        return f"<parsed file config '{super().__str__()}' at '{self.location()}'>"

    def __repr__(self):
        return super().__str__()


def _traverse_wrap(node, iter):
    for key, value in iter:
        if type(value) in recurse_handlers:
            value, iter = recurse_handlers[type(value)](value, node)
            value._key = key
            value._parent = node
            node[key] = value
            _traverse_wrap(value, iter)


class parsed_dict(dict, parsed_node):
    def merge(self, other):
        """
        Recursively merge the values of another dictionary into us.
        """
        for key, value in other.items():
            if key in self and isinstance(self[key], dict) and isinstance(value, dict):
                if not isinstance(self[key], parsed_dict):  # pragma: nocover
                    self[key] = d = parsed_dict(self[key])
                    d._key = key
                    d._parent = self
                self[key].merge(value)
            elif isinstance(value, dict):
                self[key] = d = parsed_dict(value)
                d._key = key
                d._parent = self
                _traverse_wrap(d, d.items())
            else:
                if isinstance(value, list):
                    value = parsed_list(value)
                    value._key = key
                    value._parent = self
                self[key] = value

    def rev_merge(self, other):
        """
        Recursively merge ourselves onto another dictionary.
        """
        m = parsed_dict(other)
        _traverse_wrap(m, m.items())
        m.merge(self)
        self.clear()
        self.update(m)
        for v in self.values():
            if hasattr(v, "_parent"):
                v._parent = self


class parsed_list(list, parsed_node):
    pass


def _prep_dict(node, parent):
    return parsed_dict(node), node.items()


def _prep_list(node, parent):
    return parsed_list(node), enumerate(node)


recurse_handlers = {
    dict: _prep_dict,
    parsed_dict: _prep_dict,
    list: _prep_list,
    parsed_list: _prep_list,
}


class file_ref:
    def __init__(self, node, doc, ref):
        self.node = node
        self.doc = doc
        self.ref = ref
        self.key_path = node.location()

    def resolve(self, parser, target):
        del self.node["$ref"]
        self.node.rev_merge(target)

    def __str__(self):
        return "<file ref '{}'>".format(((self.doc + "#") if self.doc else "") + self.ref)


class file_imp(file_ref):
    def __init__(self, node, doc, ref, values):
        super().__init__(node, doc, ref)
        self.values = values

    def resolve(self, parser, target):
        del self.node["$import"]
        for key in self.values:
            if key not in target:
                raise FileImportError(
                    f"'{key}' does not exist in import node '{self.ref}'"
                )
            if isinstance(target[key], dict):
                imported = parsed_dict()
                imported.merge(target[key])
                imported._key = key
                imported._parent = self.node
                if key in self.node:
                    if isinstance(self.node[key], dict):
                        imported.merge(self.node[key])
                    else:
                        warn(
                            f"Importkey '{key}' of {self} is ignored because the parent"
                            f" already contains a key '{key}'"
                            f" with value '{self.node[key]}'.",
                            ConfigurationWarning,
                            stacklevel=3,
                        )
                        continue
                self.node[key] = imported
                self._fix_references(self.node[key], parser)
            elif isinstance(target[key], list):
                imported, iter = _prep_list(target[key], self.node)
                imported._key = key
                imported._parent = self.node
                self.node[key] = imported
                self._fix_references(self.node[key], parser)
                _traverse_wrap(imported, iter)
            else:
                self.node[key] = target[key]

    def _fix_references(self, node, parser):
        # fix parser's references after the import.
        if hasattr(parser, "references"):
            for ref in parser.references:
                node_loc = node.location()
                if node_loc in ref.key_path:
                    # need to update the reference
                    # we descend the tree from the node until we reach the ref
                    # It should be here because of the merge.
                    loc_node = node
                    while node_loc != ref.key_path:
                        key = ref.key_path.split(node_loc, 1)[-1].split("/", 2)[1]
                        if key not in loc_node:  # pragma: nocover
                            raise ParserError(
                                f"Reference {ref.key_path} not found in {node_loc}. "
                                f"Should have been merged."
                            )
                        loc_node = loc_node[key]
                        node_loc += "/" + key
                    ref.node = loc_node



import abc
import functools
import os

from ..exceptions import FileReferenceError, PluginError
from ._parse_types import file_imp, file_ref, parsed_dict, recurse_handlers


class ConfigurationParser(abc.ABC):
    @abc.abstractmethod
    def parse(self, content, path=None):  # pragma: nocover
        """
        Parse configuration file content.

        :param content: str or dict content of the file to parse
        :param path: path to the file containing the configuration.
        :return: configuration tree and metadata attached as dictionaries
        """
        pass

    @abc.abstractmethod
    def generate(self, tree, pretty=False):  # pragma: nocover
        """
        Generate a string representation of the configuration tree (dictionary).

        :param dict tree: configuration tree
        :param bool pretty: if True, will add indentation to the output string
        :return: str representation of the configuration tree
        :rtype: str
        """
        pass


class ParsesReferences:
    """
    Mixin to decorate parse function of ConfigurationParser.

    Allows for imports and references inside configuration files.
    """

    def __init_subclass__(cls, **kwargs):
        super().__init_subclass__(**kwargs)
        parse = cls.parse

        def parse_with_references(self, content, path=None):
            """
            Traverse the parsed tree and resolve any `$ref` and `$import`
            """
            content, meta = parse(self, content, path)
            content = parsed_dict(content)
            self.root = content
            self.path = path or os.getcwd()
            self.is_doc = path and not os.path.isdir(path)
            self.references = []
            self.documents = {}
            self._traverse(content, content.items())
            self.resolved_documents = {}
            self._resolve_documents()
            self._resolve_references()
            return content, meta

        cls.parse = parse_with_references

    def _traverse(self, node, iter):
        # Iterates over all values in `iter` and checks for import keys, recursion or refs
        # Also wraps all nodes in their `parsed_*` counterparts.
        for key, value in iter:
            if self._is_import(key):
                self._store_import(node)
            elif type(value) in recurse_handlers:
                # The recurse handlers wrap the dicts and lists and return appropriate
                # iterators for them.
                value, iter = recurse_handlers[type(value)](value, node)
                # Set some metadata on the wrapped recursable objects.
                value._key = key
                value._parent = node
                # Overwrite the reference to the original object with a reference to the
                # wrapped object.
                node[key] = value
                # Recurse a level deeper
                self._traverse(value, iter)
            elif self._is_reference(key):
                self._store_reference(node, value)

    def _is_reference(self, key):
        return key == "$ref"

    def _is_import(self, key):
        return key == "$import"

    def _get_ref_document(self, ref, base=None):
        if "#" not in ref or ref.split("#")[0] == "":
            return None
        doc = ref.split("#")[0]
        if not os.path.isabs(doc):
            if not base:
                # reference should be relative to the current configuration file
                # to avoid recurrence issues.
                base = os.path.dirname(self.path)
            elif not os.path.isdir(base):
                base = os.path.dirname(base)
                if not os.path.exists(base):
                    raise OSError(f"Can't find reference directory '{base}'")
            doc = os.path.abspath(os.path.join(base, doc))
        return doc

    @staticmethod
    def _get_absolute_ref(node, ref):
        ref = ref.split("#")[-1]
        path = ref if ref.startswith("/") else os.path.join(node.location(), ref)
        return os.path.normpath(path).replace(os.path.sep, "/")

    def _store_reference(self, node, ref):
        # Analyzes the reference and creates a ref object from the given data
        doc = self._get_ref_document(ref, self.path)
        ref = self._get_absolute_ref(node, ref)
        if doc not in self.documents:
            self.documents[doc] = set()
        self.documents[doc].add(ref)
        self.references.append(file_ref(node, doc, ref))

    def _store_import(self, node):
        # Analyzes the import node and creates a ref object from the given data
        imp = node["$import"]
        ref = imp["ref"]
        doc = self._get_ref_document(ref)
        ref = self._get_absolute_ref(node, ref)
        if doc not in self.documents:
            self.documents[doc] = set()
        self.documents[doc].add(ref)
        if "values" not in imp:
            e = RuntimeError(f"Import node {node} is missing a 'values' list.")
            e._bsbparser_show_user = True
            raise e
        self.references.append(file_imp(node, doc, ref, imp["values"]))

    def _resolve_documents(self):
        # Iterates over the list of stored documents parses them and fetches the content
        # of each reference node.
        for file, refs in self.documents.items():
            if file is None:
                content = self.root
            else:
                from . import _try_parsers

                parser_classes = get_configuration_parser_classes()
                ext = file.split(".")[-1]
                with open(file) as f:
                    content = f.read()
                    _, content, _ = _try_parsers(content, parser_classes, ext, path=file)
            try:
                self.resolved_documents[file] = self._resolve_document(content, refs)
            except FileReferenceError as jre:
                if not file:
                    raise
                raise FileReferenceError(str(jre) + f" in document '{file}'") from None

    def _resolve_document(self, content, refs):
        resolved = {}
        for ref in refs:
            resolved[ref] = self._fetch_reference(content, ref)
        return resolved

    def _fetch_reference(self, content, ref):
        parts = [p for p in ref.split("/")[1:] if p]
        n = content
        loc = ""
        for part in parts:
            loc += "/" + part
            try:
                n = n[part]
            except KeyError:
                raise FileReferenceError(
                    f"'{loc}' in File reference '{ref}' does not exist"
                ) from None
            if not isinstance(n, dict):
                raise FileReferenceError(
                    "File references can only point to dictionaries. '{}' is a {}".format(
                        f"{loc}' in '{ref}" if loc != ref else ref,
                        type(n).__name__,
                    )
                )
        return n

    def _resolve_references(self):
        for ref in self.references:
            target = self.resolved_documents[ref.doc][ref.ref]
            ref.resolve(self, target)


@functools.cache
def get_configuration_parser_classes():
    from ..plugins import discover

    return discover("config.parsers")


def get_configuration_parser(parser, **kwargs):
    """
    Create an instance of a configuration parser that can parse configuration strings into
    configuration trees, or serialize trees into strings.

    Configuration trees can be cast into Configuration objects.
    """
    parsers = get_configuration_parser_classes()
    if parser not in parsers:
        raise PluginError(f"Configuration parser '{parser}' not found")
    return parsers[parser](**kwargs)


__all__ = [
    "ConfigurationParser",
    "ParsesReferences",
    "get_configuration_parser",
    "get_configuration_parser_classes",
]



import numpy as np

from .. import config
from ..exceptions import PackingError, PackingWarning
from ..reporting import report, warn
from ..voxels import VoxelSet
from .strategy import PlacementStrategy


class _VoxelBasedFiller:
    """
    Internal mixin for filler placement strategies.
    """

    def _fill_system(self, chunk, indicators, check_pack=True):
        voxels = VoxelSet.concatenate(
            *(p.chunk_to_voxels(chunk) for p in self.partitions)
        )
        # Define the particles for the particle system.
        particles = [
            {
                "name": name,
                # Place particles in all voxels
                "voxels": list(range(len(voxels))),
                "radius": indicator.get_radius(),
                # Indicator guesses either a global value, or a per voxel value.
                "count": indicator.guess(chunk, voxels),
            }
            for name, indicator in indicators.items()
        ]
        # Create and fill the particle system.
        system = VolumeFiller(track_displaced=False, scaffold=self.scaffold, strat=self)
        system.fill(voxels, particles, check_pack=check_pack)
        return system

    def _extract_system(self, system, chunk, indicators):
        if len(system.particles) == 0:
            return

        for pt in system.particle_types:
            cell_type = self.scaffold.cell_types[pt["name"]]
            indicator = indicators[pt["name"]]
            particle_positions = [p.position for p in system.particles if p.type is pt]
            if len(particle_positions) == 0:
                continue
            positions = np.empty((len(particle_positions), 3))
            positions[:] = particle_positions
            report(f"Placing {len(positions)} {cell_type.name} in {chunk}", level=3)
            self.place_cells(indicator, positions, chunk)


@config.node
class RandomPlacement(PlacementStrategy, _VoxelBasedFiller):
    """
    Place cells in random positions.
    """

    def place(self, chunk, indicators):
        system = self._fill_system(chunk, indicators, check_pack=True)
        self._extract_system(system, chunk, indicators)


class Particle:
    def __init__(self, radius, position):
        self.radius = radius
        self.position = position


class Neighbourhood:
    def __init__(self, epicenter, neighbours, neighbour_radius, partners, partner_radius):
        self.epicenter = epicenter
        self.neighbours = neighbours
        self.neighbour_radius = neighbour_radius
        self.partners = partners
        self.partner_radius = partner_radius

    def get_overlap(self):
        overlap = 0
        for partner in self.partners:
            for neighbour in self.neighbours:
                if partner.id == neighbour.id:
                    continue
                overlap -= min(
                    0,
                    distance(partner.position, neighbour.position)
                    - partner.radius
                    - neighbour.radius,
                )
        return overlap

    def colliding(self):
        for partner in self.partners:
            for neighbour in self.neighbours:
                if partner.id == neighbour.id:
                    continue
                if (
                    distance(partner.position, neighbour.position)
                    - partner.radius
                    - neighbour.radius
                    < 0
                ):
                    return True
        return False


class ParticleVoxel:
    def __init__(self, origin, dimensions):
        self.origin = np.array(origin)
        self.size = np.array(dimensions)


class VolumeFiller:
    def __init__(self, track_displaced=False, scaffold=None, strat=None):
        self.particle_types = []
        self.voxels = []
        self.track_displaced = track_displaced
        self.scaffold = scaffold
        self.strat = strat

    def fill(self, voxels, particles, check_pack=True):
        """
        Fill a list of voxels with Particles.

        :param bsb.voxels.VoxelSet voxels: List of voxels in which to place the particles.
        :param list[dict] particles: List of dictionary for each particle to place.
            Each dictionary needs to contain the "name" (str), "radius" (float) of
            particle.
            It should also store the "count" of particle (int | list[int]) either in total
            or for each voxel and "voxels" (list[int]) which gives in which voxel the
            placement will append.
        :param bool check_pack: If True, will check the packing factor before placing
            particles in the voxels.
        :raise PackingError: If check_pack is True and the resulting packing factor is
            greater than 0.4.
        """
        # Amount of spatial dimensions
        self.dimensions = voxels.get_raw(copy=False).shape[1]
        # Extend list of particle types in the system
        self.particle_types.extend(particles)
        # Max particle type radius
        self.max_radius = max([pt["radius"] for pt in self.particle_types])
        self.min_radius = min([pt["radius"] for pt in self.particle_types])
        # Set initial radius for collision/rearrangement to 2 times
        # the largest particle type radius
        self.search_radius = self.max_radius * 2
        # Create a list of voxels where the particles can be placed.
        self.voxels.extend(
            ParticleVoxel(ldc, size)
            for ldc, size in zip(
                voxels.as_spatial_coords(copy=False),
                voxels.get_size_matrix(copy=False),
                strict=False,
            )
        )
        if check_pack:
            pf = self.get_packing_factor()
            if self.strat is not None:
                strat_name = type(self.strat).__name__
            else:
                strat_name = "particle system"
            msg = f"Packing factor {round(pf, 2)}"
            if pf > 0.4:
                if pf > 0.64:
                    msg += " exceeds geometrical maximum packing for spheres (0.64)"
                elif pf > 0.4:
                    msg += f" too high to resolve with {strat_name}"

                count, pvol, vol = self._get_packing_factors()
                raise PackingError(
                    f"{msg}. Can not fit {round(count)} particles for a total of "
                    f"{round(pvol, 3)}μm³ micrometers into {round(vol, 3)}μm³."
                )
            elif pf > 0.2:
                warn(
                    f"{msg} is too high for good performance.",
                    PackingWarning,
                )
        # Reset particles
        self.particles = []
        for particle_type in self.particle_types:
            count = particle_type["count"]
            if count.size == 1:
                self._fill_global(particle_type)
            else:
                self._fill_per_voxel(particle_type)

    def _fill_per_voxel(self, particle_type):
        voxel_counts = particle_type["count"]
        radius = particle_type["radius"]
        if len(voxel_counts) != len(self.voxels):
            raise Exception(
                f"Particle system voxel mismatch. "
                f"Given {len(voxel_counts)} expected {len(self.voxels)}"
            )
        for voxel, count in zip(self.voxels, voxel_counts, strict=False):
            particle_type["placed"] = particle_type.get("placed", 0) + count
            placement_matrix = np.random.rand(count, self.dimensions)
            for in_voxel_pos in placement_matrix:
                particle_position = voxel.origin + in_voxel_pos * voxel.size
                self.add_particle(radius, particle_position, type=particle_type)

    def _fill_global(self, particle_type):
        particle_count = int(particle_type["count"])
        particle_type["placed"] = particle_type.get("placed", 0) + particle_count
        radius = particle_type["radius"]
        # Generate a matrix with random positions for the particles
        # Add an extra dimension to determine in which voxels to place the particles
        placement_matrix = np.random.rand(particle_count, self.dimensions + 1)
        # Generate each particle
        for row in placement_matrix:
            # Determine the voxel to be placed in.
            voxel_id = int(row[0] * len(self.voxels))
            voxel = self.voxels[voxel_id]
            # Translate the particle into the voxel based on the remaining dimensions
            particle_position = voxel.origin + row[1:] * voxel.size
            # Store the particle object
            self.add_particle(radius, particle_position, type=particle_type)

    @property
    def positions(self):
        x = np.array([p.position for p in self.particles])
        return x

    def add_particle(self, radius, position, type=None):
        particle = Particle(radius, position)
        particle.id = len(self.particles)
        particle.type = type
        self.particles.append(particle)

    def add_particles(self, radius, positions, type=None):
        for position in positions:
            self.add_particle(radius, position, type=type)

    def remove_particles(self, particles_id):
        # Remove particles with a certain id
        for index in sorted(particles_id, reverse=True):
            del self.particles[index]

    def get_packing_factor(self, particles=None, volume=None):
        """
        Calculate the packing factor of the volume where particles will be placed. It
        corresponds to the ratio of the sum of the particles' volume over the volume
        itself.

        :param list[bsb.placement.particle.Particle] | None particles: List of Particle
            to place. If None, it will use the ParticleSystem particle_types list.
        :param float | None volume: Size of the volume in which the particles will be
            placed. If None, it will use the total volume of the voxels of the
            ParticleSystem.
        :return: Packing factor.
        :rtype: float
        """
        if particles is None:
            particles_volume = sum(
                np.sum(p["count"]) * sphere_volume(p["radius"])
                for p in self.particle_types
            )
        else:
            particles_volume = sum(sphere_volume(p.radius) for p in particles)
        if volume is None:
            volume = sum(
                sum(np.prod(v.size) for v in np.array(self.voxels)[np.array(p["voxels"])])
                for p in self.particle_types
            )
        return particles_volume / volume

    def _get_packing_factors(self, particles=None, volume=None):
        if particles is None:
            particles_volume = np.sum(
                [p["count"] * sphere_volume(p["radius"]) for p in self.particle_types]
            )
            particles_count = np.sum([p["count"] for p in self.particle_types])
        else:
            particles_volume = np.sum([sphere_volume(p.radius) for p in particles])
            particles_count = len(particles)
        if volume is None:
            volume = np.sum([np.prod(v.size) for v in self.voxels])
        return particles_count, particles_volume, volume


def sphere_volume(radius):
    return 4 / 3 * np.pi * radius**3


def distance(a, b):
    return np.sqrt(np.sum((b - a) ** 2))


__all__ = ["RandomPlacement"]



from __future__ import annotations

import abc
import csv
import io
import typing
from collections import defaultdict

import numpy as np
import psutil
from tqdm import tqdm

from .. import config
from ..config import refs
from ..exceptions import ConfigurationError
from ..mixins import NotParallel
from ..storage._chunks import Chunk
from .strategy import PlacementStrategy

if typing.TYPE_CHECKING:
    from ..cell_types import CellType
    from ..storage._files import FileDependencyNode
    from ..topology.partition import Partition


@config.node
class ImportPlacement(NotParallel, PlacementStrategy, abc.ABC, classmap_entry=None):
    source: FileDependencyNode = config.file(required=True)
    cell_types: list[CellType] = config.reflist(refs.cell_type_ref, required=False)
    partitions: list[Partition] = config.reflist(refs.partition_ref, required=False)

    @config.property(default=False)
    def cache(self):
        return self.source.cache

    @cache.setter
    def cache(self, value):
        self.source.cache = bool(value)

    def place(self, chunk, indicators):
        self.parse_source(indicators)

    @abc.abstractmethod
    def parse_source(self, indicators):
        pass


@config.node
class CsvImportPlacement(ImportPlacement):
    x_header: str = config.attr(default="x")
    y_header: str = config.attr(default="y")
    z_header: str = config.attr(default="z")
    type_header: str = config.attr()
    delimiter: str = config.attr(default=",")
    progress_bar: bool = config.attr(type=bool, default=True)

    def __boot__(self):
        if not self.type_header and len(self.get_considered_cell_types()) > 1:
            raise ConfigurationError(
                "Must set `type_header` to import multiple cell types from single CSV."
            )

    def parse_source(self, indicators):
        self._reset_cache()
        chunk_size = np.array(self.scaffold.network.chunk_size)
        with self.source.provide_stream() as (fp, encoding):
            text = io.TextIOWrapper(fp, encoding=encoding, newline="")
            reader = csv.reader(text)
            headers = [h.strip().lower() for h in next(reader)]
            type_col = headers.index(self.type_header) if self.type_header else None
            coord_cols = [
                *map(headers.index, (self.x_header, self.y_header, self.z_header))
            ]
            other_cols = [
                r for r in range(len(headers)) if r not in coord_cols and r != type_col
            ]
            self._other_colnames = [headers[c] for c in other_cols]
            cts = self.get_considered_cell_types()
            if len(cts) == 1:
                name = cts[0].name
            if self.progress_bar:
                reader = tqdm(reader, desc="imported", unit=" lines")
            for i, line in enumerate(reader):
                ct_cache = self._cache[line[type_col] if type_col is not None else name]
                coords = [_safe_float(line[c]) for c in coord_cols]
                others = [_safe_float(line[c]) for c in other_cols]
                cache = ct_cache[tuple(coords // chunk_size)]
                cache[0].append(coords)
                cache[1].append(others)

                if i % 10000 == 0:
                    est_memsize = (len(others) + 3) * i * 8
                    av_mem = psutil.virtual_memory().available
                    if est_memsize > av_mem / 10:
                        print(
                            "FLUSHING, AVAILABLE MEM:",
                        )
                        self._flush(indicators)
            self._flush(indicators)

    def get_considered_cell_types(self):
        return self.cell_types or self.scaffold.cell_types.values()

    def _reset_cache(self):
        self._cache = {
            ct.name: defaultdict(lambda: [[], []])
            for ct in self.get_considered_cell_types()
        }

    def _flush(self, indicators):
        cell_types = self.get_considered_cell_types()
        iter = zip(cell_types, self._cache.values(), strict=False)
        if self.progress_bar:
            iter = tqdm(
                iter,
                desc="cell types",
                total=len(cell_types),
            )
        for ct, chunked_cache in iter:
            inner = ((Chunk(c, None), data) for c, data in chunked_cache.items())
            if self.progress_bar:
                inner = tqdm(
                    inner,
                    desc="saved",
                    total=len(chunked_cache),
                    bar_format=(
                        "{l_bar}{bar} [ {n_fmt}/{total_fmt} time left: {remaining}, "
                        "time spent: {elapsed}]"
                    ),
                )
            for chunk, data in inner:
                additional = np.array(data[1], dtype=float).T
                self.place_cells(
                    indicators[ct.name],
                    np.array(data[0]),
                    chunk,
                    additional={
                        name: col
                        for name, col in zip(
                            self._other_colnames, additional, strict=False
                        )
                    },
                )
        self._reset_cache()


def _safe_float(value: typing.Any) -> float:
    try:
        return float(value)
    except ValueError:
        return float("nan")


__all__ = ["CsvImportPlacement", "ImportPlacement"]



import functools
import importlib
import inspect
import os
import sys
import types
import warnings
from collections import defaultdict
from re import sub

import errr

from .._package_spec import warn_missing_packages
from .._util import get_qualified_class_name
from ..exceptions import (
    BootError,
    CastError,
    ConfigurationError,
    DynamicClassError,
    DynamicClassInheritanceError,
    DynamicObjectNotFoundError,
    PluginError,
    RequirementError,
    UnfitClassCastError,
    UnresolvedClassCastError,
)
from ..reporting import warn
from ._hooks import overrides


def _has_own_init(meta_subject, kwargs):
    try:
        determined_class = meta_subject.__new__.class_determinant(meta_subject, kwargs)
        return overrides(determined_class, "__init__", mro=True)
    except Exception:
        return overrides(meta_subject, "__init__", mro=True)


def make_metaclass(cls):
    # We make a `NodeMeta` class for each decorated node class, in compliance with any
    # metaclasses they might already have (to prevent metaclass confusion).
    # The purpose of the metaclass is to rewrite `__new__` and `__init__` arguments,
    # and to always call `__new__` and `__init__` in the same manner.
    # The metaclass makes it so that there are 3 overloaded constructor forms:
    #
    # MyNode({ <config dict values> })
    # MyNode(example="attr", values="here")
    # ParentNode(me=MyNode(...))
    #
    # The third makes it that type handling and other types of casting opt out early
    # and keep the object reference that the user gives them
    class ConfigArgRewrite:
        def __call__(meta_subject, *args, _parent=None, _key=None, **kwargs):
            has_own_init = _has_own_init(meta_subject, kwargs)
            # Rewrite the arguments
            primer = args[0] if args else None
            if isinstance(primer, meta_subject):
                _set_pk(primer, _parent, _key)
                return primer
            elif isinstance(primer, dict):
                args = args[1:]
                primed = primer.copy()
                primed.update(kwargs)
                kwargs = primed
            elif primer is not None and not has_own_init:
                # If we're dealing with a typical config node, the primer should be a dict
                # or already precast node. If it is not, we consider it invalid input,
                # unless the user has specified its own `__init__` function and will deal
                # with the input arguments there.
                raise ValueError(f"Unexpected positional argument '{primer}'")
            # Call the base class's new with internal arguments
            instance = meta_subject.__new__(
                meta_subject, *args, _parent=_parent, _key=_key, **kwargs
            )
            instance._config_pos_init = getattr(instance, "_config_pos_init", False)
            # Call the end user's __init__ with the rewritten arguments, if one is defined
            if has_own_init:
                sig = inspect.signature(instance.__init__)
                try:
                    # Check whether the arguments match the signature. We use `sig.bind`
                    # so that the function isn't actually called, as this could mask
                    # `TypeErrors` that occur inside the function.
                    sig.bind(*args, **kwargs)
                except TypeError as e:
                    # Since the user might not know where all these additional arguments
                    # are coming from, inform them that config nodes get passed their
                    # config attrs, and how to correctly override __init__.
                    Param = inspect.Parameter
                    help_params = {"self": Param("self", Param.POSITIONAL_OR_KEYWORD)}
                    help_params.update(sig._parameters)
                    help_params["kwargs"] = Param("kwargs", Param.VAR_KEYWORD)
                    sig._parameters = help_params
                    raise TypeError(
                        f"`{instance.__init__.__module__}.__init__` {e}."
                        + " When overriding `__init__` on config nodes, do not define"
                        + " any positional arguments, and catch any additional"
                        + " configuration attributes that are passed as keyword arguments"
                        + f": e.g. 'def __init__{sig}'"
                    ) from None
                else:
                    instance._config_pos_init = bool(len(args))
                    instance.__init__(*args, **kwargs)
            return instance

    # Avoid metaclass conflicts by prepending our rewrite class to existing metaclass MRO
    class NodeMeta(ConfigArgRewrite, *cls.__class__.__mro__):
        def __new__(cls, *args, **kwargs):
            rcls = super().__new__(cls, *args, **kwargs)
            # `__init_subclass__` refused to be called with correct subclass, so call
            # it ourselves.
            if hasattr(rcls.__bases__[0], "_cfgnode_replaced_ics"):
                rcls.__bases__[0]._cfgnode_replaced_ics(rcls, **kwargs)
            return rcls

    return NodeMeta


class NodeKwargs(dict):
    def __init__(self, instance, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.is_shortform = getattr(instance, "_config_pos_init", False)


def compose_nodes(*node_classes):
    """
    Create a composite mixin class of the given classes.

    Inherit from the returned class to inherit from more than one node class.
    """
    meta = type("ComposedMetaclass", tuple(type(cls) for cls in node_classes), {})
    return meta("CompositionMixin", node_classes, {})


def compile_class(cls):
    cls_dict = dict(cls.__dict__)
    if "__dict__" in cls_dict:
        del cls_dict["__dict__"]
    if "__weakref__" in cls_dict:
        del cls_dict["__weakref__"]
    ncls = make_metaclass(cls)(cls.__name__, cls.__bases__, cls_dict)
    for method in ncls.__dict__.values():
        _replace_closure_cells(method, cls, ncls)

    # Shitty hack, for some reason I couldn't find a way to override the first argument
    # of `__init_subclass__` methods, that would otherwise work on other classmethods,
    # so we noop the actual `__init_subclass__` and we call `__init_subclass__` ourselves
    # from the metaclass' `__new__` method, where the argument replacement works as usual.
    if (
        hasattr(ncls, "__init_subclass__")
        and "__init_subclass__" in ncls.__dict__
        and not isinstance(ncls.__init_subclass__, types.BuiltinFunctionType)
    ):
        ncls._cfgnode_replaced_ics = ncls.__init_subclass__.__func__
        ncls.__init_subclass__ = lambda *args, **kwargs: None
    classmap = getattr(ncls, "_config_dynamic_classmap", None)
    if classmap is not None:
        # Replace the reference to the old class with the new class.
        # The auto classmap entry is added in `__init_subclass__`, which happens before
        # we replace the class.
        for k, v in classmap.items():
            if v is cls:
                classmap[k] = ncls
    return ncls


def _replace_closure_cells(method, old, new):
    cl = getattr(method, "__closure__", None) or []
    for cell in cl:
        if cell.cell_contents is old:
            cell.cell_contents = new
        elif inspect.isfunction(cell.cell_contents):
            _replace_closure_cells(cell.cell_contents, old, new)


def compile_isc(node_cls, dynamic_config):
    if not dynamic_config or not dynamic_config.auto_classmap:
        return node_cls.__init_subclass__

    from ._hooks import overrides

    def dud(*args, **kwargs):
        pass

    f = node_cls.__init_subclass__ if overrides(node_cls, "__init_subclass__") else dud

    def __init_subclass__(cls, classmap_entry=MISSING, **kwargs):
        super(node_cls, cls).__init_subclass__(**kwargs)
        if classmap_entry is MISSING:
            classmap_entry = _snake_case(cls.__name__)
        if classmap_entry is not None:
            node_cls._config_dynamic_classmap[classmap_entry] = cls
        f(**kwargs)

    return classmethod(__init_subclass__)


def _snake_case(s):
    return "_".join(
        sub("([A-Z][a-z]+)", r" \1", sub("([A-Z]+)", r" \1", s.replace("-", " "))).split()
    ).lower()


def _node_determinant(cls, kwargs):
    return cls


def compile_new(node_cls, dynamic=False, pluggable=False, root=False):
    if pluggable:
        class_determinant = _get_pluggable_class
    elif dynamic:
        class_determinant = _get_dynamic_class
    else:
        class_determinant = _node_determinant

    def __new__(_cls, *args, _parent=None, _key=None, **kwargs):
        ncls = class_determinant(_cls, kwargs)
        instance = object.__new__(ncls)
        instance._config_pos_init = bool(len(args))
        _set_pk(instance, _parent, _key)
        if root:
            instance._config_isfinished = False
        instance.__post_new__(**kwargs)
        if _cls is not ncls:
            instance.__init__(*args, **kwargs)
        return instance

    __new__.class_determinant = class_determinant

    return __new__


def _set_pk(obj, parent, key):
    obj._config_parent = parent
    obj._config_key = key
    if not hasattr(obj, "_config_attr_order"):
        obj._config_attr_order = []
    if not hasattr(obj, "_config_state"):
        obj._config_state = {}
    for a in get_config_attributes(obj.__class__).values():
        if a.key:
            from ._attrs import _setattr

            _setattr(obj, a.attr_name, key)


def _missing_requirements(instance, attr, kwargs):
    # We use `self.__class__`, not `cls`, to get the proper child class.
    cls = instance.__class__
    dynamic_root = getattr(cls, "_config_dynamic_root", None)
    kwargs = NodeKwargs(instance, kwargs)
    if dynamic_root is not None:
        dynamic_attr = dynamic_root._config_dynamic_attr
        # If we are checking the dynamic attribute, but we're already a dynamic subclass,
        # we skip the required check.
        return (
            attr.attr_name == dynamic_attr
            and cls is dynamic_root
            and attr.required(kwargs)
        ) or (attr.attr_name != dynamic_attr and attr.required(kwargs))
    else:
        return attr.required(kwargs)


def compile_postnew(cls):
    def __post_new__(self, _parent=None, _key=None, **kwargs):
        attrs = get_config_attributes(self.__class__)
        self._config_attr_order = list(kwargs.keys())
        catch_attrs = [a for a in attrs.values() if hasattr(a, "__catch__")]
        leftovers = kwargs.copy()
        values = {}
        for attr in attrs.values():
            name = attr.attr_name
            value = values[name] = leftovers.pop(name, None)
            try:
                if _missing_requirements(self, attr, kwargs) and value is None:
                    raise RequirementError(f"Missing required attribute '{name}'")
            except RequirementError as e:
                # Catch both our own and possible `attr.required` RequirementErrors
                # and set the node detail before passing it on
                e.node = self
                raise
        for attr in attrs.values():
            name = attr.attr_name
            if attr.key and attr.attr_name not in kwargs:
                # If this is a "key" attribute, and the user didn't overwrite it,
                # set the attribute to the config key
                setattr(self, name, self._config_key)
                attr.flag_pristine(self)
            elif (value := values[name]) is None:
                if _is_settable_attr(attr):
                    setattr(self, name, attr.get_default())
                attr.flag_pristine(self)
            else:
                setattr(self, name, value)
                attr.flag_dirty(self)
        for key, value in leftovers.items():
            try:
                _try_catch_attrs(self, catch_attrs, key, value)
            except UncaughtAttributeError:
                try:
                    setattr(self, key, value)
                except AttributeError:
                    raise AttributeError(
                        f"Configuration attribute key '{key}' conflicts with"
                        + f" readonly class attribute on `{self.__class__.__module__}"
                        + f".{self.__class__.__name__}`."
                    ) from None
                raise ConfigurationError(f"Unknown attribute:  '{key}'") from None

    return __post_new__


def wrap_root_postnew(post_new):
    def __post_new__(self, *args, _parent=None, _key=None, _store=None, **kwargs):
        if not hasattr(self, "_meta"):
            self._meta = {"path": None, "produced": True}

        try:
            # Root node bootstrapping sequence
            _bootstrap_components(kwargs.get("components", []), file_store=_store)
            warn_missing_packages(kwargs.get("packages", []))
        except Exception as e:
            raise BootError("Failed to bootstrap configuration.") from e

        try:
            with warnings.catch_warnings(record=True) as log:
                try:
                    post_new(self, *args, _parent=None, _key=None, **kwargs)
                except (CastError, RequirementError) as e:
                    _bubble_up_exc(e, self._meta)
                self._config_isfinished = True
                _resolve_references(self)
        finally:
            _bubble_up_warnings(log)

    return __post_new__


def _is_settable_attr(attr):
    return not hasattr(attr, "fget") or attr.fset


def _bubble_up_exc(exc, meta):
    if hasattr(exc, "node") and exc.node is not None:
        node = " in " + exc.node.get_node_name()
    else:
        node = ""
    attr = f".{exc.attr}" if hasattr(exc, "attr") and exc.attr else ""
    errr.wrap(type(exc), exc, append=node + attr)


def _bubble_up_warnings(log):
    for w in log:
        m = w.message
        if hasattr(m, "node"):
            # Unpack the inner Warning that was passed instead of the warning msg
            attr = f".{m.attr.attr_name}" if hasattr(m, "attr") else ""
            warn(str(m) + " in " + m.node.get_node_name() + attr, type(m), stacklevel=4)
        else:
            warn(str(m), w.category, stacklevel=4)


def _bootstrap_components(components, file_store=None):
    from ..storage._files import CodeDependencyNode

    for component in components:
        component_node = CodeDependencyNode(component)
        component_node.file_store = file_store
        component_node.load_object()


def get_config_attributes(cls):
    attrs = {}
    if not isinstance(cls, type):
        cls = cls.__class__
    for p_cls in reversed(cls.__mro__):
        if hasattr(p_cls, "_config_attrs"):
            attrs.update(p_cls._config_attrs)
        else:
            # Scrape for mixin config attributes
            from ._attrs import ConfigurationAttribute

            attrs.update(
                {
                    key: attr
                    for key, attr in p_cls.__dict__.items()
                    if isinstance(attr, ConfigurationAttribute)
                }
            )
        for unset in getattr(p_cls, "_config_unset", []):
            attrs.pop(unset, None)
    return attrs


def _get_node_name(self):
    name = ".<missing>"
    if getattr(self, "attr_name", None) is not None:
        name = "." + str(self.attr_name)
    if getattr(self, "_config_key", None) is not None:
        name = "." + str(self._config_key)
    if hasattr(self, "_config_index"):
        if self._config_index is None:
            return "{removed}"
        else:
            name = "[" + str(self._config_index) + "]"
    if getattr(self, "name", None) is not None:
        name = "." + self.name
    if getattr(self, "_config_parent", None):
        return self._config_parent.get_node_name() + name
    else:
        return "{standalone}" + name


def make_get_node_name(node_cls, root):
    if root:
        node_cls.get_node_name = lambda self: r"{root}"
    else:
        node_cls.get_node_name = _get_node_name


class UncaughtAttributeError(Exception):
    pass


def _try_catch_attrs(node, catchers, key, value):
    # See if any of the attributes in the node can catch the value of an unknown key in
    # the configuration section. If none of them catch the value, raise an
    # `UncaughtAttributeError`
    for attr in catchers:
        try:
            _try_catch(attr.__catch__, node, key, value)
            break
        except UncaughtAttributeError:
            pass
    else:
        raise UncaughtAttributeError()


def _try_catch(catch, node, key, value):
    try:
        return catch(node, key, value)
    except Exception:
        raise UncaughtAttributeError() from None


def _get_dynamic_class(node_cls, kwargs):
    if node_cls is not node_cls._config_dynamic_root:
        # When the node is already a subclass of its dynamic root, we don't need to cast
        # it anymore.
        return node_cls

    attr_name = node_cls._config_dynamic_attr
    dynamic_attr = getattr(node_cls, attr_name)
    if attr_name in kwargs:
        loaded_cls_name = kwargs[attr_name]
    elif dynamic_attr.required(kwargs):
        raise RequirementError(f"Dynamic node must contain a '{attr_name}' attribute")
    elif dynamic_attr.should_call_default():  # pragma: nocover
        loaded_cls_name = dynamic_attr.default()
    else:
        # Fall back to the default value, or the current class.
        loaded_cls_name = dynamic_attr.default or node_cls.__name__
    module_path = ["__main__", node_cls.__module__]
    classmap = get_classmap(node_cls)
    interface = node_cls._config_dynamic_root
    try:
        dynamic_cls = _load_class(
            loaded_cls_name, module_path, interface=interface, classmap=classmap
        )
    except DynamicClassInheritanceError:
        mapped_class_msg = _get_mapped_class_msg(loaded_cls_name, classmap)
        raise UnfitClassCastError(
            f"'{loaded_cls_name}'{mapped_class_msg} is not a valid class as it does not"
            f" inherit from {node_cls.__name__}"
        ) from None
    except DynamicClassError:
        mapped_class_msg = _get_mapped_class_msg(loaded_cls_name, classmap)
        raise UnresolvedClassCastError(
            f"Could not resolve '{loaded_cls_name}'{mapped_class_msg} to a class"
        ) from None
    return dynamic_cls


def _get_pluggable_class(node_cls, kwargs):
    plugin_label = node_cls._config_plugin_name or node_cls.__name__
    if node_cls._config_plugin_key not in kwargs:
        raise CastError(
            f"Pluggable node must contain a '{node_cls._config_plugin_key}' attribute "
            f"to select a {plugin_label}"
        )
    plugin_name = kwargs[node_cls._config_plugin_key]
    plugins = node_cls.__plugins__()
    if plugin_name not in plugins:
        raise PluginError(f"Unknown {plugin_label} '{plugin_name}'")
    plugin_cls = plugins[plugin_name]
    # TODO: Enforce class inheritance
    return plugin_cls


def _get_mapped_class_msg(loaded_cls_name, classmap):
    if classmap and loaded_cls_name in classmap:
        return f" (mapped to '{classmap[loaded_cls_name]}')"
    else:
        return ""


def _load_class(cfg_classname, module_path, interface=None, classmap=None):
    if classmap and cfg_classname in classmap:
        cfg_classname = classmap[cfg_classname]
    if inspect.isclass(cfg_classname):
        class_ref = cfg_classname
        class_name = cfg_classname.__name__
    else:
        class_ref = _load_object(cfg_classname, module_path)
        class_name = class_ref.__name__

    def qualname(cls):
        return cls.__module__ + "." + cls.__name__

    if interface and not issubclass(class_ref, interface):
        raise DynamicClassInheritanceError(
            f"Dynamic class '{class_name}' must derive from {qualname(interface)}"
        )
    return class_ref


def _load_object(object_path, module_path):
    class_parts = object_path.split(".")
    object_name = class_parts[-1]
    module_name = ".".join(class_parts[:-1])
    if not module_name:
        object_ref = _search_module_path(object_name, module_path, object_path)
    else:
        object_ref = _get_module_object(object_name, module_name, object_path)

    return object_ref


def _search_module_path(class_name, module_path, cfg_classname):
    for module_name in module_path:
        module_dict = sys.modules[module_name].__dict__
        if class_name in module_dict:
            return module_dict[class_name]
    raise DynamicObjectNotFoundError("Class not found: " + cfg_classname)


def _get_module_object(object_name, module_name, object_path):
    sys.path.append(os.getcwd())
    try:
        module_ref = importlib.import_module(module_name)
    finally:
        tmp = list(reversed(sys.path))
        tmp.remove(os.getcwd())
        sys.path = list(reversed(tmp))
    try:
        return getattr(module_ref, object_name)
    except Exception:
        raise DynamicObjectNotFoundError(f"'{object_path}' not found.") from None


def make_dictable(node_cls):
    def __contains__(self, attr):
        return attr in get_config_attributes(self.__class__)

    def __getitem__(self, attr):
        if attr in get_config_attributes(self.__class__):
            return getattr(self, attr)
        else:
            raise KeyError(attr)

    def __iter__(self):
        return (attr for attr in get_config_attributes(self.__class__))

    node_cls.__contains__ = __contains__
    node_cls.__getitem__ = __getitem__
    node_cls.__iter__ = __iter__


def make_tree(node_cls):
    def get_tree(instance):
        if hasattr(instance, "__inv__") and not getattr(instance, "_config_inv", None):
            instance._config_inv = True
            inv = instance.__inv__()
            instance._config_inv = False
            return inv
        attrs = get_config_attributes(instance.__class__)
        catch_attrs = [a for a in attrs.values() if hasattr(a, "__catch__")]
        tree = {}
        for name in instance._config_attr_order:
            if name in attrs:
                attr = attrs[name]
                value = attr.tree(instance) if attr.is_dirty(instance) else None
            else:
                for catcher in catch_attrs:
                    if catcher.contains(instance, name):
                        value = catcher.tree_callback(instance, name)
                        break
                else:
                    value = getattr(instance, name, None)
            if value is not None:
                tree[name] = value
        return tree

    node_cls.__tree__ = get_tree


def make_copyable(node_cls):
    def loc_copy(instance, memo=None):
        return type(instance)(instance.__tree__())

    node_cls.__copy__ = loc_copy
    node_cls.__deepcopy__ = loc_copy


def walk_node_attributes(node):
    """
    Walk over all the child configuration nodes and attributes of ``node``.

    :returns: attribute, node, parents
    :rtype: Tuple[:class:`~.config.ConfigurationAttribute`, Any, Tuple]
    """
    attrs = get_config_attributes(node)
    if not attrs:
        if hasattr(node, "_config_attr"):
            attrs = _get_walkable_iterator(node)
        else:
            return
    for attr in attrs.values():
        yield node, attr
        # Yield but don't follow references.
        if hasattr(attr, "__ref__"):
            continue
        child = attr.__get__(node, node.__class__)
        yield from walk_node_attributes(child)


def walk_nodes(node):
    """
    Walk over all of the child configuration nodes of ``node``.

    :returns: node generator
    :rtype: Any
    """
    if hasattr(node.__class__, "_config_attrs"):
        attrs = node.__class__._config_attrs
    elif hasattr(node, "_config_attr"):
        attrs = _get_walkable_iterator(node)
    else:
        return
    yield node
    for attr in attrs.values():
        # Yield but don't follow references.
        if hasattr(attr, "__ref__"):
            continue
        child = attr.__get__(node, node.__class__)
        yield from walk_nodes(child)


def walk_node_values(start_node):
    for node, attr in walk_node_attributes(start_node):
        yield node, attr.attr_name, attr.__get__(node, node.__class__)


def _resolve_references(root, start=None, /):
    from ._attrs import _setattr

    if start is None:
        start = root
    if root._config_isfinished:
        for node, attr in walk_node_attributes(root):
            if hasattr(attr, "__ref__"):
                ref = attr.__ref__(node, root)
                _setattr(node, attr.attr_name, ref)


class WalkIterDescriptor:
    def __init__(self, n, v):
        self.attr_name = n
        self.v = v

    def __get__(self, instance, cls):
        return self.v


def _get_walkable_iterator(node):
    if isinstance(node, dict):
        walkiter = {}
        for name, value in node.items():
            walkiter[name] = WalkIterDescriptor(name, value)
        return walkiter
    elif isinstance(node, list):
        walkiter = {}
        for i, value in enumerate(node):
            walkiter[i] = WalkIterDescriptor(i, value)
        return walkiter


_classmap_registry = defaultdict(dict)


@functools.cache
def load_component_plugins():
    from ..plugins import discover

    plugins = discover("components")
    for plugin in plugins.values():
        if isinstance(plugin, dict):
            for class_name, classmap in plugin.items():
                register_classmap(class_name, classmap)

    return plugins


def register_classmap(cls_name, classmap):
    _classmap_registry[cls_name].update(classmap)


def get_classmap(cls):
    load_component_plugins()
    classmap = getattr(cls, "_config_dynamic_classmap", {})
    classmap.update(_classmap_registry[get_qualified_class_name(cls)])
    return classmap


MISSING = object()



import contextlib
import traceback
import typing

from ..reporting import warn

if typing.TYPE_CHECKING:
    import neo


class SimulationResult:
    def __init__(self, simulation):
        from neo import Block

        tree = simulation.__tree__()
        with contextlib.suppress(KeyError):
            del tree["post_prepare"]
        self.block = Block(name=simulation.name, config=tree)
        self.recorders = []

    @property
    def spiketrains(self):
        return self.block.segments[0].spiketrains

    @property
    def analogsignals(self):
        return self.block.segments[0].analogsignals

    def add(self, recorder):
        self.recorders.append(recorder)

    def create_recorder(self, flush: typing.Callable[["neo.core.Segment"], None]):
        recorder = SimulationRecorder()
        recorder.flush = flush
        self.add(recorder)
        return recorder

    def flush(self):
        from neo import Segment

        segment = Segment()
        self.block.segments.append(segment)
        for recorder in self.recorders:
            try:
                recorder.flush(segment)
            except Exception:
                traceback.print_exc()
                warn("Recorder errored out!")

    def write(self, filename, mode):
        from neo import io

        io.NixIO(filename, mode=mode).write(self.block)


class SimulationRecorder:
    def flush(self, segment: "neo.core.Segment"):
        raise NotImplementedError("Recorders need to implement the `flush` function.")


__all__ = ["SimulationResult", "SimulationRecorder"]



import typing

from .. import config
from ..config import refs
from ..config._attrs import cfglist
from .component import SimulationComponent
from .parameter import Parameter

if typing.TYPE_CHECKING:
    from ..cell_types import CellType


@config.node
class CellModel(SimulationComponent):
    """
    Cell models are simulator specific representations of a cell type.
    """

    cell_type: "CellType" = config.ref(refs.cell_type_ref, key="name")
    """
    The cell type that this model represents.
    """
    parameters: cfglist[Parameter] = config.list(type=Parameter)
    """
    The parameters of the model.
    """

    def __lt__(self, other):
        try:
            return self.name < other.name
        except Exception:
            return True

    def get_placement_set(self, chunks=None):
        return self.cell_type.get_placement_set(chunks=chunks)


__all__ = ["CellModel"]



"""
Contains all of the logic required to create commands. It should always suffice to import
just this module for a user to create their own commands.

Inherit from :class:`BaseCommand` for regular CLI style commands, or from
:class:`BsbCommand` if you want more freedom in what exactly constitutes a command to the
BSB.
"""

import argparse

from ...exceptions import CommandError
from ...reporting import report


class BaseParser(argparse.ArgumentParser):
    """
    Inherits from argparse.ArgumentParser and overloads the ``error`` method so that when
    an error occurs, instead of exiting and exception is thrown.
    """

    def error(self, message):
        """
        Raise message, instead of exiting.

        :param message: Error message
        :type message: str
        """
        raise CommandError(message)


_is_root = True


class BsbCommand:
    def add_to_parser(self):
        raise NotImplementedError("Commands must implement a `add_to_parser` method.")

    def handler(self, context):
        raise NotImplementedError("Commands must implement a `handler` method.")

    def __init_subclass__(cls, parent=None, abstract=False, name=None, **kwargs):
        global _is_root
        if abstract:
            return
        if cls.add_to_parser is BsbCommand.add_to_parser:
            raise NotImplementedError("Commands must implement a `add_to_parser` method.")
        if cls.handler is BsbCommand.handler:
            raise NotImplementedError("Commands must implement a `handler` method.")
        if name is None:
            raise CommandError(f"{cls} must register a name.")
        cls.name = name
        cls._subcommands = []
        # The very first registered command will be the RootCommand for `bsb`
        if _is_root:
            _is_root = False
        else:
            if parent is None:
                parent = RootCommand
            parent._subcommands.append(cls)


class BaseCommand(BsbCommand, abstract=True):
    def add_to_parser(self, parent, context, locals, level):
        locals = locals.copy()
        locals.update(self.get_options())
        parser = parent.add_parser(self.name)
        self.add_parser_arguments(parser)
        self.add_parser_options(parser, context, locals, level)
        parser.set_defaults(handler=self.execute_handler)
        self.add_subparsers(parser, context, self._subcommands, locals, level)
        return parser

    def add_subparsers(self, parser, context, commands, locals, level):
        if len(commands) > 0:
            subparsers = parser.add_subparsers()
            for command in commands:
                c = command()
                c._parent = self
                c.add_to_parser(subparsers, context, locals, level + 1)

    def execute_handler(self, namespace, dryrun=False):
        reduced = {}
        context = namespace._context
        for k, v in namespace.__dict__.items():
            if v is None or k in ["_context", "handler"]:
                continue
            stripped = k.lstrip("_")
            level = len(k) - len(stripped)
            if stripped not in reduced or level > reduced[stripped][0]:
                reduced[stripped] = (level, v)

        namespace.__dict__ = {k: v[1] for k, v in reduced.items()}
        self.add_locals(context)
        context.set_cli_namespace(namespace)
        report(f"Context: {context}", level=4)
        if not dryrun:
            self.handler(context)

    def add_locals(self, context):
        # Merge our options into the context, preserving those in the context as we're
        # going up the tree towards lower priority and less specific options.
        options = self.get_options()
        options.update(context.options)
        context.options = options
        if hasattr(self, "_parent"):
            self._parent.add_locals(context)

    def add_parser_options(self, parser, context, locals, level):
        merged = {}
        merged.update(context.options)
        merged.update(locals)
        for option in merged.values():
            option.add_to_parser(parser, level)

    def get_options(self):
        raise NotImplementedError(
            "BaseCommands must implement a `get_options(self)` method."
        )

    def add_parser_arguments(self, parser):
        raise NotImplementedError(
            "BaseCommands must implement an `add_parser_arguments(self, parser)` method."
        )


class RootCommand(BaseCommand, name="bsb"):
    def handler(self, context):
        pass

    def get_parser(self, context):
        parser = BaseParser()
        parser.set_defaults(_context=context)
        parser.set_defaults(handler=self.execute_handler)
        locals = self.get_options()
        self.add_parser_options(parser, context, locals, 0)
        self.add_subparsers(parser, context, self._subcommands, locals, 0)
        return parser

    def get_options(self):
        return {}


def load_root_command():
    from ...plugins import discover

    # Simply discovering the plugin modules should append them to their parent command
    # class using the `__init_subclass__` function.
    discover("commands")
    return RootCommand()


__all__ = ["BaseCommand", "BsbCommand", "RootCommand", "load_root_command"]



import contextlib as _ctxlib
import functools as _ft
import inspect as _inspect
import itertools as _it
import os as _os
import sys as _sys
import typing as _t

import numpy as _np

ichain = _it.chain.from_iterable


def merge_dicts(a, b):
    """
    Merge 2 dictionaries and their subdictionaries.
    """
    for key in b:
        if key in a and isinstance(a[key], dict) and isinstance(b[key], dict):
            merge_dicts(a[key], b[key])
        else:
            a[key] = b[key]
    return a


def obj_str_insert(__str__):
    """
    Decorator to insert the return value of __str__ into '<classname {returnvalue} at
    0x...>'.
    """

    @_ft.wraps(__str__)
    def wrapper(self):
        obj_str = object.__repr__(self)
        return obj_str.replace("at 0x", f"{__str__(self)} at 0x")

    return wrapper


@_ctxlib.contextmanager
def suppress_stdout():
    """
    Context manager that attempts to silence regular stdout and stderr.

    Some binary components may yet circumvene this if they access the underlying OS's
    stdout directly, like streaming to `/dev/stdout`.
    """
    with open(_os.devnull, "w") as devnull:
        old_stdout = _sys.stdout
        old_stderr = _sys.stderr
        _sys.stdout = devnull
        _sys.stderr = devnull
        try:
            yield
        finally:
            _sys.stdout = old_stdout
            _sys.stderr = old_stderr


def get_qualified_class_name(x):
    """
    Return an object's module and class name.
    """
    if _inspect.isclass(x):
        return f"{x.__module__}.{str(x.__name__)}"
    return f"{x.__class__.__module__}.{str(x.__class__.__name__)}"


def listify_input(value):
    """
    Turn any non-list values into a list containing the value.

    Sequences will be
    converted to a list using `list()`, `None` will  be replaced by an empty list.
    """
    if value is None:
        return []
    if isinstance(value, str):
        return [str]
    try:
        return list(value)
    except Exception:
        return [value]


def sanitize_ndarray(arr_input, shape, dtype=None):
    """
    Convert an object to an ndarray and shape, avoiding to copy it wherever possible.
    """
    kwargs = {"copy": False}
    if dtype is not None:
        kwargs["dtype"] = dtype
    arr = _np.array(arr_input, **kwargs)
    arr.shape = shape
    return arr


def assert_samelen(*args):
    """
    Assert that all input arguments have the same length.
    """
    len_ = None
    assert all(
        ((len_ := len(arg)) if len_ is None else len(arg)) == len_ for arg in args
    ), "Input arguments should be of same length."


def immutable():
    """
    Decorator to mark a method as immutable, so that any calls to it return, and are
    performed on, a copy of the instance.
    """

    def immutable_decorator(f):
        @_ft.wraps(f)
        def immutable_action(self, *args, **kwargs):
            new_instance = self.__copy__()
            f(new_instance, *args, **kwargs)
            return new_instance

        return immutable_action

    return immutable_decorator


def unique(iter_: _t.Iterable[_t.Any]):
    """
    Return a new list containing all the unique elements of an iterator.
    """
    return [*set(iter_)]


def rotation_matrix_from_vectors(vec1, vec2):
    """
    Find the rotation matrix that aligns vec1 to vec2.

    :param vec1: A 3d "source" vector
    :param vec2: A 3d "destination" vector
    :return mat: A transform matrix (3x3) which when applied to vec1, aligns it with vec2.
    """
    if (
        _np.isnan(vec1).any()
        or _np.isnan(vec2).any()
        or not _np.any(vec1)
        or not _np.any(vec2)
    ):
        raise ValueError("Vectors should not contain nan and their norm should not be 0.")
    a = (vec1 / _np.linalg.norm(vec1)).reshape(3)
    b = (vec2 / _np.linalg.norm(vec2)).reshape(3)
    v = _np.cross(a, b)
    if any(v):  # if not all zeros then
        c = _np.dot(a, b)
        s = _np.linalg.norm(v)
        kmat = _np.array([[0, -v[2], v[1]], [v[2], 0, -v[0]], [-v[1], v[0], 0]])
        return _np.eye(3) + kmat + kmat.dot(kmat) * ((1 - c) / (s**2))
    else:
        return _np.eye(3)  # cross of all zeros only occurs on identical directions



import functools
import typing

from .. import plugins

if typing.TYPE_CHECKING:
    from . import SimulationBackendPlugin
    from .adapter import SimulatorAdapter
    from .simulation import Simulation


@functools.cache
def get_backends() -> dict[str, "SimulationBackendPlugin"]:
    backends = plugins.discover("simulation_backends")
    for backend in backends.values():
        plugins._decorate_advert(backend.Simulation, backend._bsb_entry_point)
        plugins._decorate_advert(backend.Adapter, backend._bsb_entry_point)
    return backends


@functools.cache
def get_simulation_nodes() -> dict[str, "Simulation"]:
    return {name: plugin.Simulation for name, plugin in get_backends().items()}


@functools.cache
def get_simulation_adapters() -> dict[str, "SimulatorAdapter"]:
    return {name: plugin.Adapter for name, plugin in get_backends().items()}
